#!/usr/bin/env python3

import os.path as osp
import tempfile

import numpy as np
import torch

import mmcv
from mmcv.engine import collect_results_cpu, collect_results_gpu
from mmcv.runner import get_dist_info


def np2tmp(array, temp_file_name=None, tmpdir=None):
    """Save ndarray to local numpy file.

    Args:
        array (ndarray): Ndarray to save.
        temp_file_name (str): Numpy file name. If 'temp_file_name=None', this
            function will generate a file name with tempfile.NamedTemporaryFile
            to save ndarray. Default: None.
        tmpdir (str): Temporary directory to save Ndarray files. Default: None.
    Returns:
        str: The numpy file name.
    """

    if temp_file_name is None:
        temp_file_name = tempfile.NamedTemporaryFile(
            suffix=".npy", delete=False, dir=tmpdir
        ).name
    np.save(temp_file_name, array)
    return temp_file_name


def inference_edge(
    model,
    data_loader,
    out_dir=None,
    save=False,
    vis_edge=False,
    vis_dir=None,
    rescale=True,
):
    if save:
        mmcv.mkdir_or_exist(out_dir)

    if vis_edge:
        assert save, "need to have `save=True` in order to visualize"
        mmcv.mkdir_or_exist(vis_dir)

    model.eval()
    dataset = data_loader.dataset
    prog_bar = mmcv.ProgressBar(len(dataset))
    # The pipeline about how the data_loader retrieval samples from dataset:
    # sampler -> batch_sampler -> indices
    # The indices are passed to dataset_fetcher to get data from dataset.
    # data_fetcher -> collate_fn(dataset[index]) -> data_sample
    # we use batch_sampler to get correct data idx
    loader_indices = data_loader.batch_sampler

    for batch_indices, data in zip(loader_indices, data_loader):
        with torch.no_grad():
            _, edges = model(
                return_loss=False,
                return_edge=True,
                rescale=rescale,
                **data,
            )
            assert isinstance(
                edges, list
            ), f"edges should be a list but got {type(edges)}"

        img_metas = data["img_metas"][0].data[0]

        assert len(edges) == len(img_metas)

        if save:
            for edge, img_meta in zip(edges, img_metas):  # NOTE: edges
                model.module.show_edge_result(
                    edges=edge,
                    out_dir=out_dir,
                    out_prefix=osp.splitext(img_meta["ori_filename"])[0],
                    beautify=False,
                    palette=dataset.PALETTE,
                )
                if vis_edge:
                    # save visualized edge if needed
                    model.module.show_edge_result(
                        edges=edge,
                        out_dir=vis_dir,
                        out_prefix=osp.splitext(img_meta["ori_filename"])[0],
                        beautify=True,
                        palette=dataset.PALETTE,
                        beautify_threshold=0.7,  # TODO: which value should we use?
                    )

        batch_size = len(edges)
        for _ in range(batch_size):
            prog_bar.update()


def single_gpu_edge_test(
    model,
    data_loader,
    out_dir=None,
    pre_eval=False,
    save=False,
    vis_edge=False,
    vis_dir=None,
    beautify_threshold=0.7,
):
    if save:
        mmcv.mkdir_or_exist(out_dir)

    if vis_edge:
        assert save, "need to have `save=True` in order to visualize"
        mmcv.mkdir_or_exist(vis_dir)

    model.eval()
    results = []
    dataset = data_loader.dataset
    prog_bar = mmcv.ProgressBar(len(dataset))
    # The pipeline about how the data_loader retrieval samples from dataset:
    # sampler -> batch_sampler -> indices
    # The indices are passed to dataset_fetcher to get data from dataset.
    # data_fetcher -> collate_fn(dataset[index]) -> data_sample
    # we use batch_sampler to get correct data idx
    loader_indices = data_loader.batch_sampler

    for batch_indices, data in zip(loader_indices, data_loader):
        with torch.no_grad():
            _, edges = model(
                return_loss=False,
                return_edge=True,
                **data,
            )
            assert isinstance(
                edges, list
            ), f"edges should be a list but got {type(edges)}"

        img_metas = data["img_metas"][0].data[0]

        assert len(edges) == len(img_metas)

        if save:
            for edge, img_meta in zip(edges, img_metas):
                model.module.show_result(
                    edges=edge,
                    out_dir=out_dir,
                    out_prefix=osp.splitext(img_meta["ori_filename"])[0],
                    beautify=False,
                    palette=dataset.PALETTE,
                )
                if vis_edge:
                    # save visualized edge if needed
                    model.module.show_result(
                        edges=edge,
                        out_dir=vis_dir,
                        out_prefix=osp.splitext(img_meta["ori_filename"])[0],
                        beautify=True,
                        palette=dataset.PALETTE,
                        beautify_threshold=beautify_threshold,
                    )

        if pre_eval:
            # technically not edge but result
            edges = dataset.pre_edge_eval(edges, batch_indices)

        results.extend(edges)

        batch_size = len(edges)
        for _ in range(batch_size):
            prog_bar.update()

    return results


def multi_gpu_edge_test(
    model,
    data_loader,
    pre_eval=False,
    tmpdir=None,
    gpu_collect=False,
):
    model.eval()
    results = []
    dataset = data_loader.dataset
    # The pipeline about how the data_loader retrieval samples from dataset:
    # sampler -> batch_sampler -> indices
    # The indices are passed to dataset_fetcher to get data from dataset.
    # data_fetcher -> collate_fn(dataset[index]) -> data_sample
    # we use batch_sampler to get correct data idx

    # batch_sampler based on DistributedSampler, the indices only point to data
    # samples of related machine.
    loader_indices = data_loader.batch_sampler

    rank, world_size = get_dist_info()
    if rank == 0:
        prog_bar = mmcv.ProgressBar(len(dataset))

    for batch_indices, data in zip(loader_indices, data_loader):
        with torch.no_grad():
            # `return_edge` returns a tuple of (seg, edge)
            _, result = model(return_loss=False, rescale=True, return_edge=True, **data)

        if pre_eval:
            result = dataset.pre_edge_eval(result, indices=batch_indices)

        results.extend(result)

        if rank == 0:
            batch_size = len(result) * world_size
            for _ in range(batch_size):
                prog_bar.update()

    # collect results from all ranks
    if gpu_collect:
        results = collect_results_gpu(results, len(dataset))
    else:
        results = collect_results_cpu(results, len(dataset), tmpdir)
    return results



#!/usr/bin/env python3

"""Base Dataset Classes for Joint Semantic Segmentation and Edge Detection."""

import os.path as osp
import warnings
from abc import ABCMeta, abstractmethod
from collections import OrderedDict

import numpy as np
from torch.utils.data import Dataset
from prettytable import PrettyTable

import mmcv
from mmcv.utils import print_log
from mmseg.core import eval_metrics, intersect_and_union, pre_eval_to_metrics
from blette.core.metrics import (
    eval_metrics as edge_metrics,
    pre_calc as pre_edge_calc,
    pre_eval_to_metrics as pre_edge_eval_to_metrics,
)
from blette.datasets.pipelines import Compose

from potato.core.evaluation.f_boundary import (
    eval_mask_boundary,
    eval_mask_boundary_batch,
    pre_eval_to_boundary,
)


class BaseDataset(Dataset, metaclass=ABCMeta):
    """BaseDataset for joint semantic segmentation and edge detection"""

    CLASSES = None
    PALETTE = None

    img_suffix = None
    seg_map_suffix = None
    img_infos = None
    gt_seg_loader = None
    gt_edge_loader = None

    def __init__(
        self,
        pipeline,
        img_dir,
        ann_dir=None,  # seg_dir
        img_suffix=".png",
        seg_map_suffix=".png",
        split=None,
        data_root=None,
        test_mode=False,
        ignore_index=255,
        reduce_zero_label=False,
        classes=None,
        palette=None,
    ):
        self.pipeline = Compose(pipeline)
        self.img_dir = img_dir
        self.ann_dir = ann_dir
        self.img_suffix = img_suffix
        self.seg_map_suffix = seg_map_suffix

        self.split = split
        self.data_root = data_root
        self.test_mode = test_mode
        self.ignore_index = ignore_index
        self.reduce_zero_label = reduce_zero_label

        self.label_map = None
        self.CLASSES, self.PALETTE = self.get_classes_and_palette(classes, palette)

        if test_mode:
            assert (
                self.CLASSES is not None
            ), "`cls.CLASSES` or `classes` should be specified when testing"

        if self.data_root is not None:
            if not osp.isabs(self.img_dir):
                self.img_dir = osp.join(self.data_root, self.img_dir)
            if not (self.ann_dir is None or osp.isabs(self.ann_dir)):
                self.ann_dir = osp.join(self.data_root, self.ann_dir)
            if not (self.split is None or osp.isabs(self.split)):
                self.split = osp.join(self.data_root, self.split)

    def __len__(self):
        """Total number of samples of data."""
        return len(self.img_infos)

    def get_classes_and_palette(self, classes=None, palette=None):
        """Get class names of current dataset.

        Args:
            classes (Sequence[str] | str | None): If classes is None, use
                default CLASSES defined by builtin dataset. If classes is a
                string, take it as a file name. The file contains the name of
                classes where each line contains one class name. If classes is
                a tuple or list, override the CLASSES defined by the dataset.
            palette (Sequence[Sequence[int]]] | np.ndarray | None):
                The palette of segmentation map. If None is given, random
                palette will be generated. Default: None
        """
        if classes is None:
            self.custom_classes = False
            return self.CLASSES, self.PALETTE

        self.custom_classes = True
        if isinstance(classes, str):
            # take it as a file path
            class_names = mmcv.list_from_file(classes)
        elif isinstance(classes, (tuple, list)):
            class_names = classes
        else:
            raise ValueError(f"Unsupported type {type(classes)} of classes.")

        if self.CLASSES:
            if not set(class_names).issubset(self.CLASSES):
                raise ValueError("classes is not a subset of CLASSES.")

            # dictionary, its keys are the old label ids and its values
            # are the new label ids.
            # used for changing pixel labels in load_annotations.
            self.label_map = {}
            for i, c in enumerate(self.CLASSES):
                if c not in class_names:
                    self.label_map[i] = -1
                else:
                    self.label_map[i] = class_names.index(c)

        palette = self.get_palette_for_custom_classes(class_names, palette)

        return class_names, palette

    def get_palette_for_custom_classes(self, class_names, palette=None):

        if self.label_map is not None:
            # return subset of palette
            palette = []
            for old_id, new_id in sorted(self.label_map.items(), key=lambda x: x[1]):
                if new_id != -1:
                    palette.append(self.PALETTE[old_id])
            palette = type(self.PALETTE)(palette)

        elif palette is None:
            if self.PALETTE is None:
                # Get random state before set seed, and restore
                # random state later.
                # It will prevent loss of randomness, as the palette
                # may be different in each iteration if not specified.
                # See: https://github.com/open-mmlab/mmdetection/issues/5844
                state = np.random.get_state()
                np.random.seed(42)
                # random palette
                palette = np.random.randint(0, 255, size=(len(class_names), 3))
                np.random.set_state(state)
            else:
                palette = self.PALETTE

        return palette

    def get_ann_info(self, idx):
        """Get annotation by index.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Annotation info of specified index.
        """

        return self.img_infos[idx]["ann"]

    @abstractmethod
    def load_annotations(self):
        """Placeholder for loading annotations"""
        pass

    @abstractmethod
    def pre_pipeline(self, results):
        """Prepare results dict for edge pipeline."""
        pass

    def prepare_train_img(self, idx):
        """Get training data and annotations after pipeline.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Training data and annotation after pipeline with new keys
                introduced by pipeline.
        """

        img_info = self.img_infos[idx]
        ann_info = self.get_ann_info(idx)
        results = dict(img_info=img_info, ann_info=ann_info)
        self.pre_pipeline(results)
        return self.pipeline(results)

    def prepare_test_img(self, idx):
        """Get testing data after pipeline.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Testing data after pipeline with new keys introduced by
                pipeline.
        """

        img_info = self.img_infos[idx]
        results = dict(img_info=img_info)
        self.pre_pipeline(results)
        return self.pipeline(results)

    def __getitem__(self, idx):
        """Get training/test data after pipeline.

        Args:
            idx (int): Index of data.

        Returns:
            dict: Training/test data (with annotation if `test_mode` is set
                False).
        """

        if self.test_mode:
            return self.prepare_test_img(idx)
        else:
            return self.prepare_train_img(idx)

    def format_results(self, results, imgfile_prefix, indices=None, **kwargs):
        """Place holder to format result to dataset specific output."""
        raise NotImplementedError

    def get_gt_seg_map_by_idx(self, index):
        """Get one ground truth segmentation map for evaluation."""
        ann_info = self.get_ann_info(index)
        results = dict(ann_info=ann_info)
        self.pre_pipeline(results)
        self.gt_seg_loader(results)
        return results["gt_semantic_seg"]

    def get_gt_seg_maps(self, efficient_test=None):
        """Get ground truth segmentation maps for evaluation."""
        if efficient_test is not None:
            warnings.warn(
                "DeprecationWarning: ``efficient_test`` has been deprecated "
                "since potato v0.16, the ``get_gt_seg_maps()`` is CPU memory "
                "friendly by default. "
            )

        for idx in range(len(self)):
            ann_info = self.get_ann_info(idx)
            results = dict(ann_info=ann_info)
            self.pre_pipeline(results)
            self.gt_seg_loader(results)
            yield results["gt_semantic_seg"]

    def get_gt_edge_map_by_idx(self, index):
        ann_info = self.get_ann_info(index)
        results = dict(ann_info=ann_info)
        self.pre_pipeline(results)
        self.gt_edge_loader(results)
        return results["gt_semantic_edge"]

    def get_gt_edge_maps(self):
        for idx in range(len(self)):
            ann_info = self.get_ann_info(idx)
            results = dict(ann_info=ann_info)
            self.pre_pipeline(results)
            self.gt_edge_loader(results)
            yield results["gt_semantic_edge"]

    def pre_eval(self, preds, indices):
        """Collect eval result from each iteration.

        Args:
            preds (list[torch.Tensor] | torch.Tensor): the segmentation logit
                after argmax, shape (N, H, W).
            indices (list[int] | int): the prediction related ground truth
                indices.

        Returns:
            list[torch.Tensor]: (area_intersect, area_union, area_prediction,
                area_ground_truth).
        """
        # In order to compat with batch inference
        if not isinstance(indices, list):
            indices = [indices]
        if not isinstance(preds, list):
            preds = [preds]

        pre_eval_results = []

        for pred, index in zip(preds, indices):
            seg_map = self.get_gt_seg_map_by_idx(index)
            pre_eval_results.append(
                intersect_and_union(
                    pred,
                    seg_map,
                    len(self.CLASSES),
                    self.ignore_index,
                    self.label_map,
                    self.reduce_zero_label,
                )
            )

        return pre_eval_results

    def evaluate(self, results, metric="mIoU", logger=None, gt_seg_maps=None, **kwargs):
        """Evaluate the dataset.

        Args:
            results (list[tuple[torch.Tensor]] | list[str]): per image pre_eval
                results or predict segmentation map for computing evaluation
                metric.
            metric (str | list[str]): Metrics to be evaluated. 'mIoU',
                'mDice' and 'mFscore' are supported.
            logger (logging.Logger | None | str): Logger used for printing
                related information during evaluation. Default: None.
            gt_seg_maps (generator[ndarray]): Custom gt seg maps as input,
                used in ConcatDataset

        Returns:
            dict[str, float]: Default metrics.
        """
        if isinstance(metric, str):
            metric = [metric]
        allowed_metrics = ["mIoU", "mDice", "mFscore"]
        if not set(metric).issubset(set(allowed_metrics)):
            raise KeyError("metric {} is not supported".format(metric))

        eval_results = {}
        # test a list of files
        if mmcv.is_list_of(results, np.ndarray) or mmcv.is_list_of(results, str):
            if gt_seg_maps is None:
                gt_seg_maps = self.get_gt_seg_maps()
            num_classes = len(self.CLASSES)
            ret_metrics = eval_metrics(
                results,
                gt_seg_maps,
                num_classes,
                self.ignore_index,
                metric,
                label_map=self.label_map,
                reduce_zero_label=self.reduce_zero_label,
            )
        # test a list of pre_eval_results
        else:
            ret_metrics = pre_eval_to_metrics(results, metric)

        # Because dataset.CLASSES is required for per-eval.
        if self.CLASSES is None:
            class_names = tuple(range(num_classes))
        else:
            class_names = self.CLASSES

        # summary table
        ret_metrics_summary = OrderedDict(
            {
                ret_metric: np.round(np.nanmean(ret_metric_value) * 100, 2)
                for ret_metric, ret_metric_value in ret_metrics.items()
            }
        )

        # each class table
        ret_metrics.pop("aAcc", None)
        ret_metrics_class = OrderedDict(
            {
                ret_metric: np.round(ret_metric_value * 100, 2)
                for ret_metric, ret_metric_value in ret_metrics.items()
            }
        )
        ret_metrics_class.update({"Class": class_names})
        ret_metrics_class.move_to_end("Class", last=False)

        # for logger
        class_table_data = PrettyTable()
        for key, val in ret_metrics_class.items():
            class_table_data.add_column(key, val)

        summary_table_data = PrettyTable()
        for key, val in ret_metrics_summary.items():
            if key == "aAcc":
                summary_table_data.add_column(key, [val])
            else:
                summary_table_data.add_column("m" + key, [val])

        print_log("per class results:", logger)
        print_log("\n" + class_table_data.get_string(), logger=logger)
        print_log("Summary:", logger)
        print_log("\n" + summary_table_data.get_string(), logger=logger)

        # each metric dict
        for key, value in ret_metrics_summary.items():
            if key == "aAcc":
                eval_results[key] = value / 100.0
            else:
                eval_results["m" + key] = value / 100.0

        ret_metrics_class.pop("Class", None)
        for key, value in ret_metrics_class.items():
            eval_results.update(
                {
                    key + "." + str(name): value[idx] / 100.0
                    for idx, name in enumerate(class_names)
                }
            )

        return eval_results

    def pre_boundary_fscore_eval(self, preds, indices, bound_th, num_proc=1):
        """Collect eval result from each iteration for boundaries

        Args:
            preds (list[torch.Tensor] | torch.Tensor): the segmentation logit
                after argmax, shape (N, H, W).
            indices (list[int] | int): the prediction related ground truth
                indices.

        Returns:
            list[torch.Tensor]: (area_intersect, area_union, area_prediction,
                area_ground_truth).
        """
        # In order to compat with batch inference
        if not isinstance(indices, list):
            indices = [indices]
        if not isinstance(preds, list):
            preds = [preds]

        pre_eval_results = []

        if num_proc > 1:
            preds = np.asarray(preds)
            seg_maps = np.asarray([self.get_gt_seg_map_by_idx(i) for i in indices])
            Fpc, Fc = eval_mask_boundary_batch(
                preds,
                seg_maps,
                num_classes=len(self.CLASSES),
                num_proc=num_proc,
                bound_th=bound_th,
            )
            pre_eval_results.append((Fpc, Fc))
        else:
            for pred, index in zip(preds, indices):
                seg_map = self.get_gt_seg_map_by_idx(index)
                Fpc, Fc = eval_mask_boundary(
                    pred,  # already numpy
                    seg_map,  # already numpy
                    num_classes=len(self.CLASSES),
                    bound_th=bound_th,
                )
                pre_eval_results.append((Fpc, Fc))

        return pre_eval_results

    def evaluate_boundary_fscore(self, results, logger=None, **kwargs):

        fscores = pre_eval_to_boundary(results)

        class_names = self.CLASSES

        # summary table
        fscore_summary = OrderedDict({"fscore": np.round(np.nanmean(fscores) * 100, 2)})

        # each class table
        fscore_class = OrderedDict({"fscore": np.round(fscores * 100, 2)})
        fscore_class.update({"Class": class_names})
        fscore_class.move_to_end("Class", last=False)

        # for logger
        class_table_data = PrettyTable()
        for key, val in fscore_class.items():
            class_table_data.add_column(key, val)

        summary_table_data = PrettyTable()
        for key, val in fscore_summary.items():
            if key == "aAcc":
                summary_table_data.add_column(key, [val])
            else:
                summary_table_data.add_column("m" + key, [val])

        print_log("per class results:", logger)
        print_log("\n" + class_table_data.get_string(), logger=logger)
        print_log("Summary:", logger)
        print_log("\n" + summary_table_data.get_string(), logger=logger)

        # each metric dict
        eval_results = {}
        for key, value in fscore_summary.items():
            eval_results["m" + key] = value / 100.0

        fscore_class.pop("Class", None)
        for key, value in fscore_class.items():
            eval_results.update(
                {
                    key + "." + str(name): value[idx] / 100.0
                    for idx, name in enumerate(class_names)
                }
            )

        return eval_results

    def pre_edge_eval(self, preds, indices):
        """Collect eval result from each iteration.

        Args:
            preds (list[torch.Tensor] | torch.Tensor): the segmentation logit
                after argmax, shape (N, H, W).
            indices (list[int] | int): the prediction related ground truth
                indices.

        Returns:
            list[torch.Tensor]: (area_intersect, area_union, area_prediction,
                area_ground_truth).
        """
        # In order to compat with batch inference
        if not isinstance(indices, list):
            indices = [indices]
        if not isinstance(preds, list):
            preds = [preds]

        pre_eval_results = []

        for pred, index in zip(preds, indices):
            edge_map = self.get_gt_edge_map_by_idx(index)
            pre_eval_results.append(
                pre_edge_calc(
                    pred,
                    edge_map,
                    len(self.CLASSES),
                    thresh=0.7,
                )
            )

        return pre_eval_results

    def evaluate_edge(
        self,
        results,
        metric="Fscore",
        gt_edge_maps=None,
        logger=None,
        **kwargs,
    ):
        """Evaluate the dataset.

        Args:
            results (list[tuple[torch.Tensor]] | list[str]): per image pre_eval
                results or predict segmentation map for computing evaluation
                metric.
        Returns:
            dict[str, float]: Default metrics.
        """
        if isinstance(metric, str):
            metric = [metric]
        allowed_metrics = ["Fscore"]
        if not set(metric).issubset(set(allowed_metrics)):
            raise KeyError("metric {} is not supported".format(metric))

        eval_results = {}
        if mmcv.is_list_of(results, np.ndarray) or mmcv.is_list_of(results, str):
            if gt_edge_maps is None:
                gt_edge_maps = self.get_gt_edge_maps()
            num_classes = len(self.CLASSES)
            ret_metrics = edge_metrics(
                results,
                gt_edge_maps,
                num_classes=num_classes,
                metrics=metric,
                thresh=0.7,
            )
        else:
            ret_metrics = pre_edge_eval_to_metrics(results, metric)

        # summary table
        ret_metrics_summary = OrderedDict(
            {
                ret_metric: np.round(np.nanmean(ret_metric_value) * 100, 2)
                for ret_metric, ret_metric_value in ret_metrics.items()
            }
        )

        # each class table
        ret_metrics.pop("aAcc", None)
        ret_metrics_class = OrderedDict(
            {
                ret_metric: np.round(ret_metric_value * 100, 2)
                for ret_metric, ret_metric_value in ret_metrics.items()
            }
        )
        ret_metrics_class.update({"Class": self.CLASSES})
        ret_metrics_class.move_to_end("Class", last=False)

        # for logger
        class_table_data = PrettyTable()
        for key, val in ret_metrics_class.items():
            class_table_data.add_column(key, val)

        summary_table_data = PrettyTable()
        for key, val in ret_metrics_summary.items():
            if key == "aAcc":
                summary_table_data.add_column(key, [val])
            else:
                summary_table_data.add_column("m" + key, [val])

        print_log("per class results:", logger)
        print_log("\n" + class_table_data.get_string(), logger=logger)
        print_log("Summary:", logger)
        print_log("\n" + summary_table_data.get_string(), logger=logger)

        # each metric dict
        for key, value in ret_metrics_summary.items():
            if key == "aAcc":
                eval_results[key] = value / 100.0
            else:
                eval_results["m" + key] = value / 100.0

        ret_metrics_class.pop("Class", None)
        for key, value in ret_metrics_class.items():
            eval_results.update(
                {
                    key + "." + str(name): value[idx] / 100.0
                    for idx, name in enumerate(self.CLASSES)
                }
            )

        return eval_results



#!/usr/bin/env python3

"""Boundary Losses

Alexey Bokhovkin et al., Boundary Loss for Remote Sensing Imagery Semantic Segmentation
https://arxiv.org/abs/1905.07852
"""

import torch
import torch.nn as nn
import torch.nn.functional as F

from ..builder import SEG_LOSSES


def one_hot(label, num_classes):
    # FIXME: probably not efficient, but works for now
    # needed to remove ignore label which is usually 255
    onehot = []
    for i in range(num_classes):
        # this operation is probably slow on gpu since it's Long
        onehot.append(label == i)
    return torch.stack(onehot, dim=1).contiguous()


@SEG_LOSSES.register_module()
class BoundaryLoss(nn.Module):
    def __init__(
        self,
        theta0: int = 3,
        theta: int = 5,
        loss_weight=1.0,
        loss_name="loss_boundary_loss",
    ):
        super(BoundaryLoss, self).__init__()
        self.theta0 = theta0
        self.theta = theta
        self.loss_weight = loss_weight
        self._loss_name = loss_name

    def forward(self, cls_score, label, **kwargs):
        """
        Args:
            cls_score: segmentation logits (N, C, H, W)
            label: GT segmentation (N, H, W)
        """

        n, c, _, _ = cls_score.shape

        # softmax so that predicted map can be distributed in [0, 1]
        pred = torch.softmax(cls_score, dim=1)

        # one-hot vector of ground truth
        one_hot_gt = one_hot(label, c).float()
        # one_hot_gt = one_hot_gt.permute(0, 3, 1, 2).float()

        # boundary map
        gt_b = F.max_pool2d(
            1 - one_hot_gt,
            kernel_size=self.theta0,
            stride=1,
            padding=(self.theta0 - 1) // 2,
        )
        gt_b -= 1 - one_hot_gt

        pred_b = F.max_pool2d(
            1 - pred, kernel_size=self.theta0, stride=1, padding=(self.theta0 - 1) // 2
        )
        pred_b -= 1 - pred

        # extended boundary map
        gt_b_ext = F.max_pool2d(
            gt_b, kernel_size=self.theta, stride=1, padding=(self.theta - 1) // 2
        )

        pred_b_ext = F.max_pool2d(
            pred_b, kernel_size=self.theta, stride=1, padding=(self.theta - 1) // 2
        )

        # reshape
        gt_b = gt_b.view(n, c, -1)
        pred_b = pred_b.view(n, c, -1)
        gt_b_ext = gt_b_ext.view(n, c, -1)
        pred_b_ext = pred_b_ext.view(n, c, -1)

        # Precision, Recall
        P = torch.sum(pred_b * gt_b_ext, dim=2) / (torch.sum(pred_b, dim=2) + 1e-7)
        R = torch.sum(pred_b_ext * gt_b, dim=2) / (torch.sum(gt_b, dim=2) + 1e-7)

        # Boundary F1 Score
        BF1 = 2 * P * R / (P + R + 1e-7)

        return self.loss_weight * torch.mean(1 - BF1) / c

    @property
    def loss_name(self):
        return self._loss_name



import numpy as np


def beautify_edge(
    edges,
    palette,
    beautify_threshold=0.5,
):
    assert 0 < beautify_threshold < 1

    n, h, w = edges.shape
    out = np.zeros((h, w, 3))
    edges = np.where(edges >= beautify_threshold, 1, 0).astype(bool)
    edge_sum = np.zeros((h, w))

    for i in range(n):
        color = palette[i]
        edge = edges[i, :, :]
        edge_sum = edge_sum + edge
        for c in range(3):
            out[:, :, c] = np.where(edge == 1, out[:, :, c] + color[c], out[:, :, c])

    edge_sum = np.array([edge_sum, edge_sum, edge_sum])
    edge_sum = np.transpose(edge_sum, (1, 2, 0))
    idx = edge_sum > 0
    out[idx] = out[idx] / edge_sum[idx]
    out[~idx] = 255

    edges = out.astype(np.uint8)
    assert edges.ndim == 3, f"ERR: should not be {edges.ndim}dim output"
    return edges



#!/usr/bin/env python3

import torch
import torch.nn as nn

from mmcv.cnn import ConvModule, DepthwiseSeparableConvModule, build_activation_layer
from mmseg.ops import resize
from mmseg.models.decode_heads.sep_aspp_head import DepthwiseSeparableASPPModule
from blette.models.utils import LocationAdaptiveLearner

from potato.models.builder import HEADS
from potato.models.utils import maskgrad
from .maskgrad_decode_head import BaseMaskGradManyEdgeHead


@HEADS.register_module()
class TSJSBHead(BaseMaskGradManyEdgeHead):
    def __init__(
        self,
        dilations=(1, 6, 12, 18),
        c1_in_channels=256,
        c1_channels=48,
        edge_key="fuse",
        log_edge_keys=["fuse"],
        binary_edge_keys=[],
        multilabel_edge_keys=["fuse", "side5"],
        loss_binary_edge=None,
        loss_multilabel_edge=dict(type="MultiLabelEdgeLoss", loss_weight=1.0),
        mask_grad=True,
        loss_mask_edge=None,
        **kwargs,
    ):
        super(TSJSBHead, self).__init__(
            input_transform="multiple_select",
            edge_key=edge_key,
            log_edge_keys=log_edge_keys,
            binary_edge_keys=binary_edge_keys,
            multilabel_edge_keys=multilabel_edge_keys,
            loss_binary_edge=loss_binary_edge,
            loss_multilabel_edge=loss_multilabel_edge,
            mask_grad=mask_grad,
            loss_mask_edge=loss_mask_edge,
            **kwargs,
        )

        # DFF-like side outputs
        self.side_r0 = ConvModule(
            in_channels=self.in_channels[0],
            out_channels=1,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=None,
            act_cfg=None,
        )
        self.side_r1 = ConvModule(
            in_channels=self.in_channels[1],
            out_channels=1,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=None,
            act_cfg=None,
        )
        self.side_r2 = ConvModule(
            in_channels=self.in_channels[2],
            out_channels=1,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=None,
            act_cfg=None,
        )
        self.side_r4 = ConvModule(
            in_channels=self.in_channels[4],
            out_channels=self.num_classes,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=None,
            act_cfg=None,
        )
        self.weight_r4 = ConvModule(
            in_channels=self.in_channels[4],
            out_channels=self.num_classes * 4,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=None,
        )

        # adaptive weight learner
        self.ada_learner = LocationAdaptiveLearner(
            self.num_classes * 4,
            self.num_classes * 4,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        self.image_pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(
                self.in_channels[4],
                self.channels,
                1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )
        self.aspp_modules = DepthwiseSeparableASPPModule(
            in_channels=self.in_channels[4],
            channels=self.channels,
            dilations=dilations,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.c1_bottleneck = ConvModule(
            in_channels=c1_in_channels,
            out_channels=c1_channels,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        # TODO: which is better?
        # fuse in bottleneck
        # fuse in sep_bottleneck

        # fused bottleneck
        self.bottleneck = ConvModule(
            (len(dilations) + 1) * self.channels,
            self.channels,
            kernel_size=3,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.sep_bottleneck = nn.Sequential(
            DepthwiseSeparableConvModule(
                in_channels=self.channels + c1_channels + self.num_classes,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
            DepthwiseSeparableConvModule(
                in_channels=self.channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )

        self.activate = build_activation_layer(self.act_cfg)

        if mask_grad:
            self.grouped_edge_conv = nn.Conv2d(
                in_channels=self.num_classes * 2,
                out_channels=self.num_classes,
                kernel_size=1,
                groups=self.num_classes,
            )

    def forward(self, inputs):
        """Forward function"""

        # x = self._transform_inputs(inputs)  # out: list
        x = [i for i in inputs]
        # [stem, layer1, layer2, layer3, layer4, input_image]
        # h: 1/2, 1/4, 1/8, 1/8, 1/8, 1

        assert isinstance(x, list)
        bs, c, h, w = x[-1].shape
        resize_to = (h, w)  # TODO: might be too large

        s0 = resize(  # (B, 1, H, W)
            self.side_r0(x[0]),
            size=resize_to,
            mode="bilinear",
            align_corners=self.align_corners,
        )
        s1 = resize(  # (B, 1, H, W)
            self.side_r1(x[1]),
            size=resize_to,
            mode="bilinear",
            align_corners=self.align_corners,
        )
        s2 = resize(  # (B, 1, H, W)
            self.side_r2(x[2]),
            size=resize_to,
            mode="bilinear",
            align_corners=self.align_corners,
        )
        s4_out = self.side_r4(x[4])
        s4 = resize(  # (B, 19, H, W)
            s4_out,
            size=resize_to,
            mode="bilinear",
            align_corners=self.align_corners,
        )
        s4_w = resize(  # (B, 19*4, H, W)
            self.weight_r4(x[4]),
            size=resize_to,
            mode="bilinear",
            align_corners=self.align_corners,
        )

        # refine aspp with edge features
        aspp_outs = [
            resize(
                self.image_pool(x[4]),
                size=x[4].size()[2:],
                mode="bilinear",
                align_corners=self.align_corners,
            )
        ]
        # aspp_outs.append(s4)  # add edge feature (s4_out)
        aspp_outs.extend(self.aspp_modules(x[4]))
        aspp_outs = torch.cat(aspp_outs, dim=1)

        output = self.bottleneck(aspp_outs)
        c1_output = self.c1_bottleneck(x[1])
        output = resize(
            input=output,
            size=c1_output.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )
        s4_out = resize(
            # input=self.activate(s4_out),  # activation here
            input=s4_out,  # activation here
            size=c1_output.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )

        # fuse backbone and edge features
        output = torch.cat([output, c1_output, s4_out], dim=1)
        output = self.sep_bottleneck(output)

        # convert output
        output = self.cls_seg(output)

        # semantic edge detection
        # s4_w = self.activate(s4_w)  # NOTE: activation here (reduces metrics)
        ada_weights = self.ada_learner(s4_w)  # (B, 19, 4, H, W)
        slice4 = s4[:, 0:1, :, :]  # (B, 1, H, W)

        fused_edges = torch.cat((slice4, s0, s1, s2), 1)
        for i in range(s4.size(1) - 1):
            slice4 = s4[:, i + 1 : i + 2, :, :]  # (B, 1, H, W)
            fused_edges = torch.cat(
                (fused_edges, slice4, s0, s1, s2), 1
            )  # (B, 19*4, H, W)

        fused_edges = fused_edges.view(  # (B, 19, H, W)
            fused_edges.size(0),
            self.num_classes,
            -1,
            fused_edges.size(2),
            fused_edges.size(3),
        )
        # fused_edges = torch.mul(fused_edges, torch.sigmoid(ada_weights))
        fused_edges = torch.mul(fused_edges, ada_weights)
        fused_edges = torch.sum(fused_edges, 2)  # (B, 19, H, W)

        if self.mask_grad:
            # calculate mask gradient
            seg_out = resize(
                output,  # we want logits
                size=resize_to,
                mode="bilinear",
                align_corners=self.align_corners,
            )
            mask_edge = maskgrad(
                seg_map=seg_out,
                sensitivity=100,
            )
            # sliced concatenation and k-grouped convolution
            hold = []
            for c in range(fused_edges.size(1)):
                hold.append(fused_edges[:, c, ...].unsqueeze(1))
                hold.append(mask_edge[:, c, ...].unsqueeze(1))
            edge = torch.cat(hold, dim=1)
            edge = self.grouped_edge_conv(edge)  # logits
            return output, dict(fuse=edge, side4=s4), mask_edge
        else:
            return output, dict(fuse=fused_edges, side4=s4)



#!/usr/bin/env python3

import numpy as np
from PIL import Image


def apply_mask(image, mask, color):
    """Apply the given mask to the image."""
    for c in range(3):
        image[:, :, c] = np.where(mask == 1, image[:, :, c] + color[c], image[:, :, c])
    return image


def visualize_prediction(dataset, path, pred):
    n, h, w = pred.shape
    image = np.zeros((h, w, 3))
    # image = image.astype(np.uint32)

    # FIXME: check if the colors are correct and the same with original dataset

    if dataset == "cityscapes":
        colors = [
            [128, 64, 128],
            [244, 35, 232],
            [70, 70, 70],
            [102, 102, 156],
            [190, 153, 153],
            [153, 153, 153],
            [250, 170, 30],
            [220, 220, 0],
            [107, 142, 35],
            [152, 251, 152],
            [70, 130, 180],
            [220, 20, 60],
            [255, 0, 0],
            [0, 0, 142],
            [0, 0, 70],
            [0, 60, 100],
            [0, 80, 100],
            [0, 0, 230],
            [119, 11, 32],
        ]
    else:
        assert dataset == "sbd"
        colors = [
            [128, 0, 0],
            [0, 128, 0],
            [128, 128, 0],
            [0, 0, 128],
            [128, 0, 128],
            [0, 128, 128],
            [128, 128, 128],
            [64, 0, 0],
            [192, 0, 0],
            [64, 128, 0],
            [192, 128, 0],
            [64, 0, 128],
            [192, 0, 128],
            [64, 128, 128],
            [192, 128, 128],
            [0, 64, 0],
            [128, 64, 0],
            [0, 192, 0],
            [128, 192, 0],
            [0, 64, 128],
        ]

    pred = np.where(pred >= 0.5, 1, 0).astype(np.bool)
    edge_sum = np.zeros((h, w))

    for i in range(n):
        color = colors[i]
        edge = pred[i, :, :]
        edge_sum = edge_sum + edge
        masked_image = apply_mask(image, edge, color)

    edge_sum = np.array([edge_sum, edge_sum, edge_sum])
    edge_sum = np.transpose(edge_sum, (1, 2, 0))
    idx = edge_sum > 0
    masked_image[idx] = masked_image[idx] / edge_sum[idx]
    masked_image[~idx] = 255

    # io.imsave(path, masked_image / 255)

    masked_image = masked_image.astype(np.uint8)
    img = Image.fromarray(masked_image)
    img.save(path)



#!/usr/bin/env python3

import os.path as osp
import warnings

import torch.distributed as dist
from torch.nn.modules.batchnorm import _BatchNorm

from mmcv.runner import DistEvalHook
from mmcv.runner import EvalHook


class EvalSegHook(EvalHook):
    """Single GPU EvalHook, with efficient test support.

    Args:
        by_epoch (bool): Determine perform evaluation by epoch or by iteration.
            If set to True, it will perform by epoch. Otherwise, by iteration.
            Default: False.
        efficient_test (bool): Whether save the results as local numpy files to
            save CPU memory during evaluation. Default: False.
        pre_eval (bool): Whether to use progressive mode to evaluate model.
            Default: False.
    Returns:
        list: The prediction results.
    """

    greater_keys = ["mIoU", "mAcc", "aAcc"]

    def __init__(
        self, *args, by_epoch=False, efficient_test=False, pre_eval=False, **kwargs
    ):
        super().__init__(*args, by_epoch=by_epoch, **kwargs)
        self.pre_eval = pre_eval
        self.latest_results = None

        if efficient_test:
            warnings.warn(
                "DeprecationWarning: ``efficient_test`` for evaluation hook "
                "is deprecated, the evaluation hook is CPU memory friendly "
                "with ``pre_eval=True`` as argument for ``single_gpu_test()`` "
                "function"
            )

    def _do_evaluate(self, runner):
        """perform evaluation and save ckpt."""
        if not self._should_evaluate(runner):
            return

        from potato.apis import single_gpu_seg_test

        results = single_gpu_seg_test(
            runner.model, self.dataloader, show=False, pre_eval=self.pre_eval
        )
        self.latest_results = results

        runner.log_buffer.clear()
        runner.log_buffer.output["eval_iter_num"] = len(self.dataloader)
        key_score = self.evaluate(runner, results)
        if self.save_best:
            self._save_ckpt(runner, key_score)

    def evaluate(self, runner, results):
        """Evaluate the results.
        Args:
            runner (:obj:`mmcv.Runner`): The underlined training runner.
            results (list): Output results.
        """
        eval_res = self.dataloader.dataset.evaluate(
            results, logger=runner.logger, **self.eval_kwargs
        )

        for name, val in eval_res.items():
            runner.log_buffer.output[name] = val
        runner.log_buffer.mode = "val"  # patch
        runner.log_buffer.ready = True

        if self.save_best is not None:
            # If the performance of model is pool, the `eval_res` may be an
            # empty dict and it will raise exception when `self.save_best` is
            # not None. More details at
            # https://github.com/open-mmlab/mmdetection/issues/6265.
            if not eval_res:
                warnings.warn(
                    "Since `eval_res` is an empty dict, the behavior to save "
                    "the best checkpoint will be skipped in this evaluation."
                )
                return None

            if self.key_indicator == "auto":
                # infer from eval_results
                self._init_rule(self.rule, list(eval_res.keys())[0])
            return eval_res[self.key_indicator]

        return None


class DistEvalSegHook(DistEvalHook):
    """Distributed EvalHook, with efficient test support.

    Args:
        by_epoch (bool): Determine perform evaluation by epoch or by iteration.
            If set to True, it will perform by epoch. Otherwise, by iteration.
            Default: False.
        efficient_test (bool): Whether save the results as local numpy files to
            save CPU memory during evaluation. Default: False.
        pre_eval (bool): Whether to use progressive mode to evaluate model.
            Default: False.
    Returns:
        list: The prediction results.
    """

    greater_keys = ["mIoU", "mAcc", "aAcc"]

    def __init__(
        self, *args, by_epoch=False, efficient_test=False, pre_eval=False, **kwargs
    ):
        super().__init__(*args, by_epoch=by_epoch, **kwargs)
        self.pre_eval = pre_eval
        self.latest_results = None
        if efficient_test:
            warnings.warn(
                "DeprecationWarning: ``efficient_test`` for evaluation hook "
                "is deprecated, the evaluation hook is CPU memory friendly "
                "with ``pre_eval=True`` as argument for ``multi_gpu_test()`` "
                "function"
            )

    def _do_evaluate(self, runner):
        """perform evaluation and save ckpt."""
        # Synchronization of BatchNorm's buffer (running_mean
        # and running_var) is not supported in the DDP of pytorch,
        # which may cause the inconsistent performance of models in
        # different ranks, so we broadcast BatchNorm's buffers
        # of rank 0 to other ranks to avoid this.
        if self.broadcast_bn_buffer:
            model = runner.model
            for name, module in model.named_modules():
                if isinstance(module, _BatchNorm) and module.track_running_stats:
                    dist.broadcast(module.running_var, 0)
                    dist.broadcast(module.running_mean, 0)

        if not self._should_evaluate(runner):
            return

        tmpdir = self.tmpdir
        if tmpdir is None:
            tmpdir = osp.join(runner.work_dir, ".eval_hook")

        from potato.apis import multi_gpu_seg_test

        results = multi_gpu_seg_test(
            runner.model,
            self.dataloader,
            tmpdir=tmpdir,
            gpu_collect=self.gpu_collect,
            pre_eval=self.pre_eval,
        )
        self.latest_results = results
        runner.log_buffer.clear()

        if runner.rank == 0:
            print("\n")
            runner.log_buffer.output["eval_iter_num"] = len(self.dataloader)
            key_score = self.evaluate(runner, results)

            if self.save_best:
                self._save_ckpt(runner, key_score)

    def evaluate(self, runner, results):
        """Evaluate the results.
        Args:
            runner (:obj:`mmcv.Runner`): The underlined training runner.
            results (list): Output results.
        """
        eval_res = self.dataloader.dataset.evaluate(
            results, logger=runner.logger, **self.eval_kwargs
        )

        for name, val in eval_res.items():
            runner.log_buffer.output[name] = val
        runner.log_buffer.mode = "val"  # patch
        runner.log_buffer.ready = True

        if self.save_best is not None:
            # If the performance of model is pool, the `eval_res` may be an
            # empty dict and it will raise exception when `self.save_best` is
            # not None. More details at
            # https://github.com/open-mmlab/mmdetection/issues/6265.
            if not eval_res:
                warnings.warn(
                    "Since `eval_res` is an empty dict, the behavior to save "
                    "the best checkpoint will be skipped in this evaluation."
                )
                return None

            if self.key_indicator == "auto":
                # infer from eval_results
                self._init_rule(self.rule, list(eval_res.keys())[0])
            return eval_res[self.key_indicator]

        return None



#!/usr/bin/env python3

"""
Wrap mmseg dataset to add additional functions
"""
from collections import OrderedDict

import numpy as np
from prettytable import PrettyTable

from mmcv.utils import print_log
from mmseg.datasets import CityscapesDataset

from potato.core.evaluation.f_boundary import (
    eval_mask_boundary,
    eval_mask_boundary_batch,
    pre_eval_to_boundary,
)
from .builder import DATASETS

FBOUNDARY_EVAL_DATASET_MAPPING = {
    "CityscapesDataset": "WrappedCityscapesDataset",
}


@DATASETS.register_module()
class WrappedCityscapesDataset(CityscapesDataset):
    def pre_boundary_fscore_eval(self, preds, indices, bound_th, num_proc=1):
        """Collect eval result from each iteration for boundaries

        Args:
            preds (list[torch.Tensor] | torch.Tensor): the segmentation logit
                after argmax, shape (N, H, W).
            indices (list[int] | int): the prediction related ground truth
                indices.

        Returns:
            list[torch.Tensor]: (area_intersect, area_union, area_prediction,
                area_ground_truth).
        """
        # In order to compat with batch inference
        if not isinstance(indices, list):
            indices = [indices]
        if not isinstance(preds, list):
            preds = [preds]

        pre_eval_results = []

        if num_proc > 1:
            preds = np.asarray(preds)
            seg_maps = np.asarray([self.get_gt_seg_map_by_idx(i) for i in indices])
            Fpc, Fc = eval_mask_boundary_batch(
                preds,
                seg_maps,
                num_classes=len(self.CLASSES),
                num_proc=num_proc,
                bound_th=bound_th,
            )
            pre_eval_results.append((Fpc, Fc))
        else:
            for pred, index in zip(preds, indices):
                seg_map = self.get_gt_seg_map_by_idx(index)
                Fpc, Fc = eval_mask_boundary(
                    pred,  # already numpy
                    seg_map,  # already numpy
                    num_classes=len(self.CLASSES),
                    bound_th=bound_th,
                )
                pre_eval_results.append((Fpc, Fc))

        return pre_eval_results

    def evaluate_boundary_fscore(self, results, logger=None, **kwargs):

        fscores = pre_eval_to_boundary(results)

        class_names = self.CLASSES

        # summary table
        fscore_summary = OrderedDict({"fscore": np.round(np.nanmean(fscores) * 100, 2)})

        # each class table
        fscore_class = OrderedDict({"fscore": np.round(fscores * 100, 2)})
        fscore_class.update({"Class": class_names})
        fscore_class.move_to_end("Class", last=False)

        # for logger
        class_table_data = PrettyTable()
        for key, val in fscore_class.items():
            class_table_data.add_column(key, val)

        summary_table_data = PrettyTable()
        for key, val in fscore_summary.items():
            if key == "aAcc":
                summary_table_data.add_column(key, [val])
            else:
                summary_table_data.add_column("m" + key, [val])

        print_log("per class results:", logger)
        print_log("\n" + class_table_data.get_string(), logger=logger)
        print_log("Summary:", logger)
        print_log("\n" + summary_table_data.get_string(), logger=logger)

        # each metric dict
        eval_results = {}
        for key, value in fscore_summary.items():
            eval_results["m" + key] = value / 100.0

        fscore_class.pop("Class", None)
        for key, value in fscore_class.items():
            eval_results.update(
                {
                    key + "." + str(name): value[idx] / 100.0
                    for idx, name in enumerate(class_names)
                }
            )

        return eval_results



#!/usr/bin/env python3

from abc import ABCMeta, abstractmethod

import torch
import torch.nn as nn
from mmcv.runner import BaseModule, auto_fp16, force_fp32
from mmseg.core import build_pixel_sampler
from mmseg.ops import resize

from potato.models.builder import build_seg_loss, build_edge_loss, build_joint_loss
from potato.models.losses import accuracy, edge_accuracy


class BaseMultiTaskDecodeHead(BaseModule, metaclass=ABCMeta):
    """Base class for BaseMultiTaskDecodeHead.

    The key differences are:
    - `loss_decode` is now `loss_seg`
    - added `loss_edge`, and `loss_joint`
    """

    def __init__(
        self,
        in_channels,
        channels,
        *,
        num_classes,
        binary_edge=False,
        dropout_ratio=0.1,
        conv_cfg=None,
        norm_cfg=None,
        act_cfg=dict(type="ReLU"),
        in_index=-1,
        input_transform=None,
        loss_seg=dict(type="CrossEntropyLoss", use_sigmoid=False, loss_weight=1.0),
        loss_edge=None,
        loss_joint=None,
        ignore_index=255,
        sampler=None,
        align_corners=False,
        init_cfg=dict(type="Normal", std=0.01, override=dict(name="conv_seg")),
        no_conv_seg=False,  # HACK: some models use their own convolution at the end
    ):
        super(BaseMultiTaskDecodeHead, self).__init__(init_cfg)
        self._init_inputs(in_channels, in_index, input_transform)
        self.channels = channels
        assert num_classes > 1, f"ERR: number of classes > 1: {num_classes}"
        self.num_classes = num_classes
        self.binary_edge = binary_edge
        self.dropout_ratio = dropout_ratio
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.in_index = in_index

        self.ignore_index = ignore_index
        self.align_corners = align_corners

        # setting up loss_seg
        if isinstance(loss_seg, dict):
            if "num_classes" in loss_seg.keys():
                loss_seg.update(dict(num_classes=self.num_classes))
            self.loss_seg = build_seg_loss(loss_seg)
        elif isinstance(loss_seg, (list, tuple)):
            self.loss_seg = nn.ModuleList()
            for loss in loss_seg:
                if "num_classes" in loss.keys():
                    loss.update(dict(num_classes=self.num_classes))
                self.loss_seg.append(build_seg_loss(loss))
        else:
            raise TypeError(
                f"loss_seg must be a dict or sequence of dict,\
                but got {type(loss_seg)}"
            )

        # setting up loss_edge
        if loss_edge is None:
            self.loss_edge = None
        else:
            if isinstance(loss_edge, dict):
                if "num_classes" in loss_edge.keys():
                    loss_edge.update(dict(num_classes=self.num_classes))
                self.loss_edge = build_edge_loss(loss_edge)
            elif isinstance(loss_edge, (list, tuple)):
                self.loss_edge = nn.ModuleList()
                for loss in loss_edge:
                    if "num_classes" in loss.keys():
                        loss.update(dict(num_classes=self.num_classes))
                    self.loss_edge.append(build_edge_loss(loss))
            else:
                raise TypeError(
                    f"loss_edge must be a dict or sequence of dict,\
                    but got {type(loss_joint)}"
                )

        # setting up loss_joint
        if loss_joint is None:
            self.loss_joint = None
        else:
            if isinstance(loss_joint, dict):
                if "num_classes" in loss_joint.keys():
                    loss_joint.update(dict(num_classes=self.num_classes))
                self.loss_joint = build_joint_loss(loss_joint)
            elif isinstance(loss_joint, (list, tuple)):
                self.loss_joint = nn.ModuleList()
                for loss in loss_joint:
                    if "num_classes" in loss.keys():
                        loss.update(dict(num_classes=self.num_classes))
                    self.loss_joint.append(build_joint_loss(loss))
            else:
                raise TypeError(
                    f"loss_joint must be a dict or sequence of dict,\
                    but got {type(loss_joint)}"
                )

        if sampler is not None:
            self.sampler = build_pixel_sampler(sampler, context=self)
        else:
            self.sampler = None

        if no_conv_seg:
            pass
        else:
            self.conv_seg = nn.Conv2d(channels, num_classes, kernel_size=1)
            if dropout_ratio > 0:
                self.dropout = nn.Dropout2d(dropout_ratio)
            else:
                self.dropout = None
        self.fp16_enabled = False

    def extra_repr(self):
        """Extra repr."""
        s = (
            f"input_transform={self.input_transform}, "
            f"ignore_index={self.ignore_index}, "
            f"align_corners={self.align_corners}"
        )
        return s

    def _init_inputs(self, in_channels, in_index, input_transform):
        """Check and initialize input transforms.

        The in_channels, in_index and input_transform must match.
        Specifically, when input_transform is None, only single feature map
        will be selected. So in_channels and in_index must be of type int.
        When input_transform

        Args:
            in_channels (int|Sequence[int]): Input channels.
            in_index (int|Sequence[int]): Input feature index.
            input_transform (str|None): Transformation type of input features.
                Options: 'resize_concat', 'multiple_select', None.
                'resize_concat': Multiple feature maps will be resize to the
                    same size as first one and than concat together.
                    Usually used in FCN head of HRNet.
                'multiple_select': Multiple feature maps will be bundle into
                    a list and passed into decode head.
                None: Only one select feature map is allowed.
        """

        if input_transform is not None:
            assert input_transform in ["resize_concat", "multiple_select"]
        self.input_transform = input_transform
        self.in_index = in_index
        if input_transform is not None:
            assert isinstance(in_channels, (list, tuple))
            assert isinstance(in_index, (list, tuple))
            assert len(in_channels) == len(in_index)
            if input_transform == "resize_concat":
                self.in_channels = sum(in_channels)
            else:
                self.in_channels = in_channels
        else:
            assert isinstance(in_channels, int)
            assert isinstance(in_index, int)
            self.in_channels = in_channels

    def _transform_inputs(self, inputs):
        """Transform inputs for decoder.

        Args:
            inputs (list[Tensor]): List of multi-level img features.

        Returns:
            Tensor: The transformed inputs
        """

        if self.input_transform == "resize_concat":
            inputs = [inputs[i] for i in self.in_index]
            upsampled_inputs = [
                resize(
                    input=x,
                    size=inputs[0].shape[2:],
                    mode="bilinear",
                    align_corners=self.align_corners,
                )
                for x in inputs
            ]
            inputs = torch.cat(upsampled_inputs, dim=1)
        elif self.input_transform == "multiple_select":
            inputs = [inputs[i] for i in self.in_index]
        else:
            inputs = inputs[self.in_index]

        return inputs

    @auto_fp16()
    @abstractmethod
    def forward(self, inputs, **kwargs):
        """Placeholder of forward function."""
        pass

    def forward_train(
        self,
        inputs,
        img_metas,
        gt_semantic_seg,
        gt_semantic_edge,
        train_cfg,
    ):
        # NOTE: segmentation results are logits, but edges may not
        seg_logits, edge_logits = self(inputs)

        # convert multilabel to binary edge
        if self.binary_edge and gt_semantic_edge.shape[1] != 1:
            # convert to binary
            gt_semantic_edge = (
                (torch.sum(gt_semantic_edge, axis=1) > 0).unsqueeze(1).float()
            )

        # Assume that there are always seg_losses
        losses = self.seg_losses(
            seg_logit=seg_logits,
            seg_label=gt_semantic_seg,
        )
        if self.loss_edge is not None:
            losses.update(
                self.edge_losses(
                    edge_logit=edge_logits,
                    edge_label=gt_semantic_edge,
                )
            )
        if self.loss_joint is not None:
            losses.update(
                self.joint_losses(
                    seg_logit=seg_logits,
                    edge_logit=edge_logits,
                    seg_label=gt_semantic_seg,
                    edge_label=gt_semantic_edge,
                )
            )

        return losses

    def cls_seg(self, feat):
        """Classify each pixel."""
        if self.dropout is not None:
            feat = self.dropout(feat)
        output = self.conv_seg(feat)
        return output

    def forward_test(self, inputs, img_metas, test_cfg, return_edge=False, **kwargs):
        seg, edge = self(inputs)

        if return_edge:
            return dict(seg=seg, edge=edge)
        else:
            return dict(seg=seg)

    @force_fp32(apply_to=("seg_logit"))
    def seg_losses(self, seg_logit, seg_label):
        """Compute segmentation loss."""
        loss = dict()

        seg_logit = resize(
            input=seg_logit,
            size=seg_label.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )
        if self.sampler is not None:
            seg_weight = self.sampler.sample(seg_logit, seg_label)
        else:
            seg_weight = None
        seg_label = seg_label.squeeze(1)

        if not isinstance(self.loss_seg, nn.ModuleList):
            losses_seg = [self.loss_seg]
        else:
            losses_seg = self.loss_seg
        for loss_seg in losses_seg:
            if loss_seg.loss_name not in loss:
                loss[loss_seg.loss_name] = loss_seg(
                    seg_logit,
                    seg_label,
                    weight=seg_weight,
                    ignore_index=self.ignore_index,
                )
            else:
                loss[loss_seg.loss_name] += loss_seg(
                    seg_logit,
                    seg_label,
                    weight=seg_weight,
                    ignore_index=self.ignore_index,
                )

        loss["acc_seg"] = accuracy(seg_logit, seg_label)
        return loss

    @force_fp32(apply_to=("edge_logit"))
    def edge_losses(self, edge_logit, edge_label):
        """Compute edge loss."""
        loss = dict()

        edge_logit = resize(
            input=edge_logit,
            size=edge_label.shape[2:],  # (b, cls, h, w)
            mode="bilinear",
            align_corners=self.align_corners,
        )

        if not isinstance(self.loss_edge, nn.ModuleList):
            losses_edge = [self.loss_edge]
        else:
            losses_edge = self.loss_edge
        for loss_edge in losses_edge:
            if loss_edge.loss_name not in loss:
                loss[loss_edge.loss_name] = loss_edge(
                    edge_logit,
                    edge_label,
                    ignore_index=self.ignore_index,
                )
            else:
                loss[loss_edge.loss_name] += loss_edge(
                    edge_logit,
                    edge_label,
                    ignore_index=self.ignore_index,
                )

        for metric_name, v in edge_accuracy(edge_logit, edge_label).items():
            loss["edge" + "_" + metric_name] = v

        return loss

    @force_fp32(apply_to=("seg_logit", "edge_logit"))
    def joint_losses(self, seg_logit, edge_logit, seg_label, edge_label):
        """Compute joint loss.

        Currently only supports binary edges
        """
        loss = dict()

        seg_logit = resize(
            input=seg_logit,
            size=seg_label.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )
        edge_logit = resize(
            input=edge_logit,
            size=edge_label.shape[2:],  # (b, 19, h, w)
            mode="bilinear",
            align_corners=self.align_corners,
        )
        if self.sampler is not None:
            seg_weight = self.sampler.sample(seg_logit, seg_label)
        else:
            seg_weight = None
        seg_label = seg_label.squeeze(1)

        if not isinstance(self.loss_joint, nn.ModuleList):
            losses_joint = [self.loss_joint]
        else:
            losses_joint = self.loss_joint
        for loss_joint in losses_joint:
            if loss_joint.loss_name not in loss:
                loss[loss_joint.loss_name] = loss_joint(
                    cls_score=seg_logit,
                    label=seg_label,
                    edge=edge_logit,
                    edge_label=edge_label,
                    weight=seg_weight,
                    ignore_index=self.ignore_index,
                )
            else:
                loss[loss_joint.loss_name] += loss_joint(
                    cls_score=seg_logit,
                    label=seg_label,
                    edge=edge_logit,
                    edge_label=edge_label,
                    weight=seg_weight,
                    ignore_index=self.ignore_index,
                )

        return loss



#!/usr/bin/env python3

import torch
import torch.nn as nn

from mmcv.runner import force_fp32
from mmseg.ops import resize

from potato.models.builder import HEADS
from potato.models.losses import accuracy, edge_accuracy
from .multitask_decode_head import BaseMultiTaskDecodeHead


@HEADS.register_module()
class BinaryHEDHead(BaseMultiTaskDecodeHead):
    def __init__(
        self,
        merging="attention",
        deep_supervision=True,
        **kwargs,
    ):
        super().__init__(
            input_transform="multiple_select",
            binary_edge=True,
            no_conv_seg=True,
            init_cfg=dict(type="Normal", std=0.01),
            **kwargs,
        )

        self.seg_branch = nn.ModuleList(
            [
                nn.Conv2d(
                    in_channels=i,
                    out_channels=self.num_classes,  # 2
                    kernel_size=1,
                )
                for i in self.in_channels
            ]
        )

        self.edge_branch = nn.ModuleList(
            [
                nn.Conv2d(
                    in_channels=i,
                    out_channels=1,  # 1
                    kernel_size=1,
                )
                for i in self.in_channels
            ]
        )

        self.deep_supervision = deep_supervision
        self.merging = merging
        if merging == "attention":
            self.seg_queries = nn.ModuleList(
                [
                    nn.Conv2d(
                        in_channels=i,
                        out_channels=self.num_classes,  # 2
                        kernel_size=1,
                    )
                    for i in self.in_channels
                ]
            )
            self.edge_queries = nn.ModuleList(
                [
                    nn.Conv2d(
                        in_channels=i,
                        out_channels=1,  # 1
                        kernel_size=1,
                    )
                    for i in self.in_channels
                ]
            )
        elif merging == "learned":
            self.merge_seg = nn.Conv2d(
                in_channels=self.num_classes * len(self.in_channels),
                out_channels=self.num_classes,  # 2
                kernel_size=1,
            )
            self.merge_edge = nn.Conv2d(
                in_channels=self.num_classes * len(self.in_channels),
                out_channels=1,  # 1
                kernel_size=1,
            )
        else:
            # no merging
            pass

    def forward(self, inputs):
        inputs = list(inputs)
        img = inputs.pop(-1)  # pop img
        b, _, h, w = img.shape

        # seg and edge branch
        seg_preds = []
        edge_preds = []
        for x, sconv, econv in zip(inputs, self.seg_branch, self.edge_branch):
            seg_preds.append(sconv(x))
            edge_preds.append(econv(x))

        if self.merging == "attention":
            seg_queries = torch.cat(
                [
                    resize(
                        input=q(feat),
                        size=(h, w),
                        mode="bilinear",
                        align_corners=self.align_corners,
                    )
                    for q, feat in zip(self.seg_queries, inputs)
                ],
                dim=1,
            )
            edge_queries = torch.cat(
                [
                    resize(
                        input=q(feat),
                        size=(h, w),
                        mode="bilinear",
                        align_corners=self.align_corners,
                    )
                    for q, feat in zip(self.edge_queries, inputs)
                ],
                dim=1,
            )
            segs = torch.cat(
                [
                    resize(
                        input=p,
                        size=(h, w),
                        mode="bilinear",
                        align_corners=self.align_corners,
                    )
                    for p in seg_preds
                ],
                dim=1,
            )
            edges = torch.cat(
                [
                    resize(
                        input=p,
                        size=(h, w),
                        mode="bilinear",
                        align_corners=self.align_corners,
                    )
                    for p in edge_preds
                ],
                dim=1,
            )
            seg_queries = seg_queries.reshape(b, -1, 2, h, w)
            attn = torch.softmax(seg_queries, dim=1)
            segs = segs.reshape(b, -1, 2, h, w)
            combined_seg = torch.sum(attn * segs, dim=1)

            edge_queries = edge_queries.reshape(b, -1, 1, h, w)
            attn = torch.softmax(edge_queries, dim=1)
            edges = edges.reshape(b, -1, 1, h, w)
            combined_edge = torch.sum(attn * edges, dim=1)

        elif self.merging == "learned":
            segs = torch.cat(
                [
                    resize(
                        input=p,
                        size=(h, w),
                        mode="bilinear",
                        align_corners=self.align_corners,
                    )
                    for p in seg_preds
                ],
                dim=1,
            )
            edges = torch.cat(
                [
                    resize(
                        input=p,
                        size=(h, w),
                        mode="bilinear",
                        align_corners=self.align_corners,
                    )
                    for p in edge_preds
                ],
                dim=1,
            )
            combined_seg = self.merge_seg(segs)
            combined_edge = self.merge_edge(edges)
        else:
            combined_seg = seg_preds[-1]
            combined_edge = edge_preds[-1]

        if self.deep_supervision:
            seg_preds.append(combined_seg)
            edge_preds.append(combined_edge)
            return seg_preds, edge_preds
        else:
            return combined_seg, combined_edge

    @force_fp32(apply_to=("seg_logit",))
    def multi_seg_losses(self, seg_logit, seg_label):
        """Compute segmentation loss."""
        loss = dict()

        for s in seg_logit:
            s = resize(
                input=s,
                size=seg_label.shape[2:],
                mode="bilinear",
                align_corners=self.align_corners,
            )
            if self.sampler is not None:
                seg_weight = self.sampler.sample(s, seg_label)
            else:
                seg_weight = None
            l = seg_label.squeeze(1)

            if not isinstance(self.loss_seg, nn.ModuleList):
                losses_seg = [self.loss_seg]
            else:
                losses_seg = self.loss_seg
            for loss_seg in losses_seg:
                if loss_seg.loss_name not in loss:
                    loss[loss_seg.loss_name] = loss_seg(
                        s,
                        l,
                        weight=seg_weight,
                        ignore_index=self.ignore_index,
                    )
                else:
                    loss[loss_seg.loss_name] += loss_seg(
                        s,
                        l,
                        weight=seg_weight,
                        ignore_index=self.ignore_index,
                    )

        loss["acc_seg"] = accuracy(seg_logit[-1], seg_label.squeeze(1))
        return loss

    @force_fp32(apply_to=("edge_logit",))
    def multi_edge_losses(self, edge_logit, edge_label):
        """Compute edge loss."""
        loss = dict()

        for e in edge_logit:
            e = resize(
                input=e,
                size=edge_label.shape[2:],  # (b, cls, h, w)
                mode="bilinear",
                align_corners=self.align_corners,
            )

            if not isinstance(self.loss_edge, nn.ModuleList):
                losses_edge = [self.loss_edge]
            else:
                losses_edge = self.loss_edge
            for loss_edge in losses_edge:
                if loss_edge.loss_name not in loss:
                    loss[loss_edge.loss_name] = loss_edge(
                        e,
                        edge_label,
                        ignore_index=self.ignore_index,
                    )
                else:
                    loss[loss_edge.loss_name] += loss_edge(
                        e,
                        edge_label,
                        ignore_index=self.ignore_index,
                    )

        for metric_name, v in edge_accuracy(edge_logit[-1], edge_label).items():
            loss["edge" + "_" + metric_name] = v

        return loss

    def forward_train(
        self,
        inputs,
        img_metas,
        gt_semantic_seg,
        gt_semantic_edge,
        train_cfg,
    ):
        # convert multilabel to binary edge
        if self.binary_edge and gt_semantic_edge.shape[1] != 1:
            # convert to binary
            gt_semantic_edge = (
                (torch.sum(gt_semantic_edge, axis=1) > 0).unsqueeze(1).float()
            )

        seg_logits, edge_logits = self(inputs)

        if self.deep_supervision:
            losses = self.multi_seg_losses(
                seg_logit=seg_logits,
                seg_label=gt_semantic_seg,
            )
            losses.update(
                self.multi_edge_losses(
                    edge_logit=edge_logits,
                    edge_label=gt_semantic_edge,
                )
            )
        else:
            # Assume that there are always seg_losses
            losses = self.seg_losses(
                seg_logit=seg_logits,
                seg_label=gt_semantic_seg,
            )
            losses.update(
                self.edge_losses(
                    edge_logit=edge_logits,
                    edge_label=gt_semantic_edge,
                )
            )
        return losses

    def forward_test(self, inputs, img_metas, test_cfg, return_edge=False, **kwargs):
        seg, edge = self(inputs)
        if self.deep_supervision:
            seg = seg[-1]
            edge = edge[-1]

        if return_edge:
            return dict(seg=seg, edge=edge)
        else:
            return dict(seg=seg)



#!/usr/bin/env python3

"""Reference model for GSCNN and Multi-Label GSCNN

Difference:
- Different feature fusion
- Removed sigmoid activations in the model because loss uses logits
- Added final convolution for edge prediction
- Different canny edge (image gradient)
- Shape stream has activations between gated-conv
- Depthwise Separable Convs at the end

Multi-Label GSCNN
- edge attention loss produces `nan` whle training possibily due to learning rate
    and network architecture.
"""

import torch
import torch.nn as nn

from mmcv.cnn import ConvModule, DepthwiseSeparableConvModule
from mmseg.ops import resize
from mmseg.models.decode_heads.sep_aspp_head import DepthwiseSeparableASPPModule

from potato.models.builder import HEADS
from potato.models.utils import maskgrad
from potato.ops.canny_edge import UnNormalizedCanny
from potato.ops.torch_canny_edge import TorchUnNormalizedSoftCanny
from .multitask_decode_head import BaseMultiTaskDecodeHead
from ..utils import ResNetGSCNNShapeStream


@HEADS.register_module()
class GSCNNHead(BaseMultiTaskDecodeHead):
    """GSCNN head with minor edits for stable training"""

    def __init__(
        self,
        dilations=(1, 6, 12, 18),
        c1_in_channels=256,
        c1_channels=48,
        binary_edge=True,
        cv2_canny=False,
        loss_edge=dict(type="BinaryEdgeLoss", loss_weight=1.0 * 20),
        **kwargs,
    ):
        super().__init__(
            input_transform="multiple_select",
            binary_edge=binary_edge,
            loss_edge=loss_edge,
            **kwargs,
        )

        if binary_edge:
            num_classes = 1
        else:
            num_classes = self.num_classes

        self.shape_stream = ResNetGSCNNShapeStream(
            in_channels=self.in_channels,
            num_classes=num_classes,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
            align_corners=self.align_corners,
        )
        self.cw = ConvModule(
            in_channels=num_classes + 1,  # logit + canny
            out_channels=1,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.aspp_edge_conv = ConvModule(
            in_channels=1,
            out_channels=self.channels,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.image_pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(
                self.in_channels[4],
                self.channels,
                1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )
        self.aspp_modules = DepthwiseSeparableASPPModule(
            in_channels=self.in_channels[4],
            channels=self.channels,
            dilations=dilations,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.c1_bottleneck = ConvModule(
            in_channels=c1_in_channels,
            out_channels=c1_channels,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.bottleneck = ConvModule(
            (len(dilations) + 2) * self.channels,  # (img_pool, edge, *dilations)
            self.channels,
            3,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.sep_bottleneck = nn.Sequential(
            DepthwiseSeparableConvModule(
                in_channels=self.channels + c1_channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
            DepthwiseSeparableConvModule(
                in_channels=self.channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )

        if cv2_canny:
            # cv2, fast, non-differentiable
            self.canny = UnNormalizedCanny()
        else:
            # torch, a bit slower, differentiable, no hysteresis
            self.canny = TorchUnNormalizedSoftCanny()

    def forward(self, inputs):
        """Forward function"""

        # [stem, layer1, layer2, layer3, layer4, input_image]
        # h: 1/2, 1/4, 1/8, 1/8, 1/8, 1
        x = [i for i in inputs]

        assert isinstance(x, list)
        assert len(x) == 6

        bs, c, h, w = x[-1].shape
        resize_to = (h, w)

        # Shape Stream
        edge_out = self.shape_stream(x)

        # Canny Edge Fusion
        canny = self.canny(x[-1])
        edge_feat = resize(
            torch.sigmoid(edge_out),
            size=resize_to,
            mode="bilinear",
            align_corners=self.align_corners,
        )
        edge_feat = torch.cat((edge_feat, canny), dim=1)
        edge_feat = self.cw(edge_feat)
        edge_feat = resize(
            edge_feat,
            size=x[4].size()[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )

        # ASPP
        aspp_outs = [
            resize(
                self.image_pool(x[4]),
                size=x[4].size()[2:],
                mode="bilinear",
                align_corners=self.align_corners,
            )
        ]
        aspp_outs.append(self.aspp_edge_conv(edge_feat))
        aspp_outs.extend(self.aspp_modules(x[4]))
        aspp_outs = torch.cat(aspp_outs, dim=1)
        bot_out = self.bottleneck(aspp_outs)

        # Bottlenecks
        c1_output = self.c1_bottleneck(x[1])
        bot_out = resize(
            input=bot_out,
            size=c1_output.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )
        bot_out = torch.cat([bot_out, c1_output], dim=1)
        bot_out = self.sep_bottleneck(bot_out)

        seg_out = self.cls_seg(bot_out)
        return seg_out, edge_out


@HEADS.register_module()
class MultiLabelGSCNNHead(GSCNNHead):
    """Custom GSCNN head that outputs multi-label edges

    NOTE:
    - minimal changes to the original `GSCNNHead` above!
    - also added option for mask gradient to improve the edges
    """

    def __init__(
        self,
        mask_grad=False,
        binary_edge=False,
        loss_edge=dict(type="MultiLabelEdgeLoss", loss_weight=1.0),
        **kwargs,
    ):
        super().__init__(
            binary_edge=binary_edge,
            loss_edge=loss_edge,
            **kwargs,
        )
        self.mask_grad = mask_grad
        if self.mask_grad:
            self.grouped_edge_conv = nn.Conv2d(
                in_channels=self.num_classes * 2,
                out_channels=self.num_classes,
                kernel_size=1,
                groups=self.num_classes,
            )

    def forward(self, inputs):
        seg_out, edge_out = super().forward(inputs)

        if self.mask_grad:

            # CHECK: order?
            _seg_out = resize(
                seg_out,  # we want logits
                size=edge_out.shape[2:],
                mode="bilinear",
                align_corners=self.align_corners,
            )
            mask_edge = maskgrad(
                seg_map=_seg_out,
                sensitivity=100,
            )

            # sliced concatenation and k-grouped convolution
            hold = []
            for c in range(edge_out.size(1)):
                hold.append(edge_out[:, c, ...].unsqueeze(1))
                hold.append(mask_edge[:, c, ...].unsqueeze(1))
            edge = torch.cat(hold, dim=1)
            edge_out = self.grouped_edge_conv(edge)  # logits

        return seg_out, edge_out



#!/usr/bin/env python3

import os.path as osp

import numpy as np
from PIL import Image

import mmcv
from mmcv.utils import print_log

from .builder import DATASETS
from .custom import CustomJointDataset, OTFCustomJointDataset

from pyEdgeEval.datasets.cityscapes_attributes import (
    CITYSCAPES_inst_labelIds,
    CITYSCAPES_labelIds,
    CITYSCAPES_label2trainId,
)

CITYSCAPES_CLASSES = (
    "road",
    "sidewalk",
    "building",
    "wall",
    "fence",
    "pole",
    "traffic light",
    "traffic sign",
    "vegetation",
    "terrain",
    "sky",
    "person",
    "rider",
    "car",
    "truck",
    "bus",
    "train",
    "motorcycle",
    "bicycle",
)
CITYSCAPES_PALETTE = [
    [128, 64, 128],
    [244, 35, 232],
    [70, 70, 70],
    [102, 102, 156],
    [190, 153, 153],
    [153, 153, 153],
    [250, 170, 30],
    [220, 220, 0],
    [107, 142, 35],
    [152, 251, 152],
    [70, 130, 180],
    [220, 20, 60],
    [255, 0, 0],
    [0, 0, 142],
    [0, 0, 70],
    [0, 60, 100],
    [0, 80, 100],
    [0, 0, 230],
    [119, 11, 32],
]
ignore_ids = [2, 3]


@DATASETS.register_module()
class JointCityscapesDataset(CustomJointDataset):
    CLASSES = CITYSCAPES_CLASSES
    PALETTE = CITYSCAPES_PALETTE

    def __init__(
        self,
        img_suffix="_leftImg8bit.png",
        seg_map_suffix="_gtFine_labelTrainIds.png",
        edge_map_suffix="_gtProc_raw_edge.png",
        inst_sensitive=True,
        thin=False,
        test_mode=False,
        **kwargs,
    ):

        if test_mode:
            if inst_sensitive:
                if thin:
                    edge_map_suffix = "_gtProc_thin_isedge.png"
                else:
                    edge_map_suffix = "_gtProc_raw_isedge.png"
            else:
                if thin:
                    edge_map_suffix = "_gtProc_thin_edge.png"
                else:
                    edge_map_suffix = "_gtProc_raw_edge.png"
        else:
            if inst_sensitive:
                edge_map_suffix = "_gtProc_isedge.png"
            else:
                edge_map_suffix = "_gtProc_edge.png"

        super(JointCityscapesDataset, self).__init__(
            img_suffix=img_suffix,
            seg_map_suffix=seg_map_suffix,
            edge_map_suffix=edge_map_suffix,
            inst_sensitive=inst_sensitive,
            **kwargs,
        )

    @staticmethod
    def _convert_to_label_id(result):
        """Convert trainId to id for cityscapes."""
        if isinstance(result, str):
            result = np.load(result)
        import cityscapesscripts.helpers.labels as CSLabels

        result_copy = result.copy()
        for trainId, label in CSLabels.trainId2label.items():
            result_copy[result == trainId] = label.id

        return result_copy

    def results2img(self, results, imgfile_prefix, to_label_id, indices=None):
        """Write the segmentation results to images.

        Args:
            results (list[ndarray]): Testing results of the
                dataset.
            imgfile_prefix (str): The filename prefix of the png files.
                If the prefix is "somepath/xxx",
                the png files will be named "somepath/xxx.png".
            to_label_id (bool): whether convert output to label_id for
                submission.
            indices (list[int], optional): Indices of input results,
                if not set, all the indices of the dataset will be used.
                Default: None.

        Returns:
            list[str: str]: result txt files which contains corresponding
            semantic segmentation images.
        """
        if indices is None:
            indices = list(range(len(self)))

        mmcv.mkdir_or_exist(imgfile_prefix)
        result_files = []
        for result, idx in zip(results, indices):
            if to_label_id:
                result = self._convert_to_label_id(result)
            filename = self.img_infos[idx]["filename"]
            basename = osp.splitext(osp.basename(filename))[0]

            png_filename = osp.join(imgfile_prefix, f"{basename}.png")

            output = Image.fromarray(result.astype(np.uint8)).convert("P")
            import cityscapesscripts.helpers.labels as CSLabels

            palette = np.zeros((len(CSLabels.id2label), 3), dtype=np.uint8)
            for label_id, label in CSLabels.id2label.items():
                palette[label_id] = label.color

            output.putpalette(palette)
            output.save(png_filename)
            result_files.append(png_filename)

        return result_files

    def format_results(self, results, imgfile_prefix, to_label_id=True, indices=None):
        """Format the results into dir (standard format for Cityscapes
        evaluation).

        Args:
            results (list): Testing results of the dataset.
            imgfile_prefix (str): The prefix of images files. It
                includes the file path and the prefix of filename, e.g.,
                "a/b/prefix".
            to_label_id (bool): whether convert output to label_id for
                submission. Default: False
            indices (list[int], optional): Indices of input results,
                if not set, all the indices of the dataset will be used.
                Default: None.

        Returns:
            tuple: (result_files, tmp_dir), result_files is a list containing
                the image paths, tmp_dir is the temporal directory created
                for saving json/png files when img_prefix is not specified.
        """
        if indices is None:
            indices = list(range(len(self)))

        assert isinstance(results, list), "results must be a list."
        assert isinstance(indices, list), "indices must be a list."

        result_files = self.results2img(results, imgfile_prefix, to_label_id, indices)

        return result_files

    def evaluate(self, results, metric="mIoU", logger=None, imgfile_prefix=None):
        """Evaluation in Cityscapes/default protocol.

        Args:
            results (list): Testing results of the dataset.
            metric (str | list[str]): Metrics to be evaluated.
            logger (logging.Logger | None | str): Logger used for printing
                related information during evaluation. Default: None.
            imgfile_prefix (str | None): The prefix of output image file,
                for cityscapes evaluation only. It includes the file path and
                the prefix of filename, e.g., "a/b/prefix".
                If results are evaluated with cityscapes protocol, it would be
                the prefix of output png files. The output files would be
                png images under folder "a/b/prefix/xxx.png", where "xxx" is
                the image name of cityscapes. If not specified, a temp file
                will be created for evaluation.
                Default: None.

        Returns:
            dict[str, float]: Cityscapes/default metrics.
        """

        eval_results = dict()
        metrics = metric.copy() if isinstance(metric, list) else [metric]
        if "cityscapes" in metrics:
            eval_results.update(
                self._evaluate_cityscapes(results, logger, imgfile_prefix)
            )
            metrics.remove("cityscapes")
        if len(metrics) > 0:
            eval_results.update(
                super(JointCityscapesDataset, self).evaluate(results, metrics, logger)
            )

        return eval_results

    def _evaluate_cityscapes(self, results, logger, imgfile_prefix):
        """Evaluation in Cityscapes protocol.

        Args:
            results (list): Testing results of the dataset.
            logger (logging.Logger | str | None): Logger used for printing
                related information during evaluation. Default: None.
            imgfile_prefix (str | None): The prefix of output image file

        Returns:
            dict[str: float]: Cityscapes evaluation results.
        """
        try:
            import cityscapesscripts.evaluation.evalPixelLevelSemanticLabeling as CSEval  # noqa
        except ImportError:
            raise ImportError(
                'Please run "pip install cityscapesscripts" to '
                "install cityscapesscripts first."
            )
        msg = "Evaluating in Cityscapes style"
        if logger is None:
            msg = "\n" + msg
        print_log(msg, logger=logger)

        result_dir = imgfile_prefix

        eval_results = dict()
        print_log(f"Evaluating results under {result_dir} ...", logger=logger)

        CSEval.args.evalInstLevelScore = True
        CSEval.args.predictionPath = osp.abspath(result_dir)
        CSEval.args.evalPixelAccuracy = True
        CSEval.args.JSONOutput = False

        seg_map_list = []
        pred_list = []

        # when evaluating with official cityscapesscripts,
        # **_gtFine_labelIds.png is used
        for seg_map in mmcv.scandir(
            self.ann_dir, "_gtFine_labelIds.png", recursive=True
        ):
            seg_map_list.append(osp.join(self.other_dir, seg_map))
            pred_list.append(CSEval.getPrediction(CSEval.args, seg_map))

        eval_results.update(
            CSEval.evaluateImgLists(pred_list, seg_map_list, CSEval.args)
        )

        return eval_results


@DATASETS.register_module()
class OTFJointCityscapesDataset(OTFCustomJointDataset):
    """OTF edge generation for Cityscapes dataset

    - should only be used for training the model
    """

    CLASSES = CITYSCAPES_CLASSES
    PALETTE = CITYSCAPES_PALETTE

    def __init__(
        self,
        img_suffix="_leftImg8bit.png",
        seg_map_suffix="_gtFine_labelIds.png",
        inst_map_suffix="_gtFine_instanceIds.png",
        inst_sensitive=True,  # default instance sensitive
        labelIds=CITYSCAPES_labelIds,
        inst_labelIds=CITYSCAPES_inst_labelIds,
        ignore_indicies=[2, 3],
        label2trainId=CITYSCAPES_label2trainId,
        **kwargs,
    ):
        super(OTFJointCityscapesDataset, self).__init__(
            img_suffix=img_suffix,
            seg_map_suffix=seg_map_suffix,
            inst_map_suffix=inst_map_suffix,
            inst_sensitive=inst_sensitive,
            labelIds=labelIds,
            inst_labelIds=inst_labelIds,
            ignore_indices=ignore_indicies,
            label2trainId=label2trainId,
            **kwargs,
        )

    def format_results(self, **kwargs):
        raise ValueError("ERR: Should not use OTF for test set")

    def evaluate(self, **kwargs):
        raise ValueError("ERR: Should not use OTF for evaluation!")



#!/usr/bin/env python3

from .gated_conv import GatedSpatialConv2d
from .shape_stream import ResNetGSCNNShapeStream
from .mask_grad import maskgrad

__all__ = [
    "GatedSpatialConv2d",
    "ResNetGSCNNShapeStream",
    "maskgrad",
]



#!/usr/bin/env python3

from .base import BaseJointSegDet
from .custom_encoder_decoder import CustomEncoderDecoder
from .binary_joint_encoder_decoder import BinaryJointEncoderDecoder
from .joint_encoder_decoder import JointEncoderDecoder

__all__ = [
    "BaseJointSegDet",
    "CustomEncoderDecoder",
    "BinaryJointEncoderDecoder",
    "JointEncoderDecoder",
]



#!/usr/bin/env python3

"""Canny Edge"""

from copy import deepcopy

import cv2
import numpy as np
import torch

from .transforms import UnNormalize


def canny_cv2(
    img: np.ndarray,
    thresh_low: int = 10,
    thresh_high: int = 100,
):
    """Canny Edge

    Args:
        img (np.ndarray): (batch, c, h, w), np.uint8
    """

    if img.ndim == 3:
        # add batch direction
        img = img[None, ...]
    assert img.ndim == 4

    bs = img.size(0)
    bs, h, w, c = img.shape
    out = np.zeros((bs, 1, h, w))
    for i in range(bs):
        # should the input be RGB or Gray?
        _img = cv2.cvtColor(img[i], cv2.COLOR_RGB2GRAY)

        # values are 0 ~ 255
        out[i] = cv2.Canny(_img, thresh_low, thresh_high)

    # NOTE: normalize output
    out = out / 255

    if out.ndim == 4 and out.shape[0] == 1:
        out = out.squeeze(0)

    return out


class UnNormalizedCanny(object):
    def __init__(
        self,
        low_threshold=10,
        high_threshold=100,
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        inplace=True,
    ):
        self.thresh_low = low_threshold
        self.thresh_high = high_threshold
        self.unnorm = UnNormalize(
            mean=mean,
            std=std,
            inplace=inplace,
        )

    def __call__(self, img):
        non_batch = False
        if img.ndim == 3:
            # add batch direction
            non_batch = True
            img = img.unsqueeze(0)
        assert img.ndim == 4
        bs, c, h, w = img.shape
        device = img.device

        # NOTE: need to normalize since the threashold values are larger than 1
        im_ten = self.unnorm(deepcopy(img.detach().cpu()))
        # NOTE: cllip before casting
        im_arr = im_ten.numpy().transpose((0, 2, 3, 1)).clip(0, 255).astype(np.uint8)
        canny = np.zeros((bs, 1, h, w))
        for i in range(bs):
            # should the input be RGB or Gray?
            _img = cv2.cvtColor(im_arr[i], cv2.COLOR_RGB2GRAY)
            # values are 0 ~ 255
            canny[i] = cv2.Canny(_img, self.thresh_low, self.thresh_high)
        # NOTE: normalize output
        canny = canny / 255.0
        canny = torch.from_numpy(canny).to(device).float()

        if non_batch:
            canny = canny.squeeze(0)

        return canny


class OldUnNormalizedCanny(UnNormalizedCanny):
    def __call__(self, img):
        non_batch = False
        if img.ndim == 3:
            # add batch direction
            non_batch = True
            img = img.unsqueeze(0)
        assert img.ndim == 4
        bs, c, h, w = img.shape
        device = img.device

        im_arr = img.cpu().numpy().transpose((0, 2, 3, 1)).astype(np.uint8)
        canny = np.zeros((bs, 1, h, w))
        for i in range(bs):
            # values are 0 ~ 255
            canny[i] = cv2.Canny(im_arr[i], self.thresh_low, self.thresh_high)
        canny = torch.from_numpy(canny).to(device).float()

        if non_batch:
            canny = canny.squeeze(0)

        return canny



#!/usr/bin/env python3

import torch
import torch.nn as nn

from mmcv.cnn import ConvModule


class GatedSpatialConv2d(ConvModule):
    def __init__(
        self,
        in_channels,
        out_channels,
        kernel_size=1,
        stride=1,
        padding=0,
        dilation=1,
        groups=1,
        bias=False,
        conv_cfg=None,
        norm_cfg=None,
        act_cfg=None,
        **kwargs,
    ):
        """Gated Spatial Conv2D

        Args:
            in_channels
            out_channels
            kernel_size
            stride
            padding
            dilation
            groups
            bias

        Original implementation:
        - BN -> conv2d(bias=True) -> ReLU -> conv2d(bias=True) -> BN -> sigmoid

        This implementation:
        - conv2d(bias=False) -> BN -> ReLU -> conv2d(bias=True) -> BN -> sigmoid

        This is because BN and activations are already applied to side outputs.
        """

        assert norm_cfg is not None
        assert act_cfg is not None

        super().__init__(
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=kernel_size,
            stride=stride,
            padding=padding,
            dilation=dilation,
            groups=groups,
            bias=bias,
            conv_cfg=conv_cfg,
            norm_cfg=norm_cfg,
            act_cfg=act_cfg,
            **kwargs,
        )

        self._gate_conv = nn.Sequential(
            ConvModule(
                in_channels=in_channels + 1,
                out_channels=in_channels + 1,
                kernel_size=1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
            ConvModule(
                in_channels=in_channels + 1,
                out_channels=1,
                kernel_size=1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=None,
            ),
            nn.Sigmoid(),
        )

    def forward(
        self,
        x: torch.Tensor,
        gating_features: torch.Tensor,
        activate: bool = False,  # default: don't activate after conv
        norm: bool = True,
    ) -> torch.Tensor:
        for layer in self.order:
            if layer == "conv":
                if self.with_explicit_padding:
                    x = self.padding_layer(x)
                alphas = self._gate_conv(
                    torch.cat(
                        [x, gating_features],
                        dim=1,
                    )
                )
                # residual connection
                x = x * (alphas + 1)
                x = self.conv(x)
            elif layer == "norm" and norm and self.with_norm:
                x = self.norm(x)
            elif layer == "act" and activate and self.with_activation:
                x = self.activate(x)

        return x



#!/usr/bin/env python3

from mmcv.runner import force_fp32
from mmseg.ops import resize

from potato.models.builder import build_edge_loss
from .multitask_decode_head import BaseMultiTaskDecodeHead
from .manyedge_decode_head import BaseManyEdgeDecodeHead


class BaseMaskGradMultiTaskHead(BaseMultiTaskDecodeHead):
    def __init__(
        self,
        mask_grad=False,
        loss_mask_edge=dict(type="DualityLoss", loss_weight=1.0),
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.mask_grad = mask_grad
        if self.mask_grad:
            assert isinstance(loss_mask_edge, dict)
            self.mask_edge_loss = build_edge_loss(loss_mask_edge)

    @force_fp32(apply_to=("edge_logit", "maskgrad_prob"))
    def maskgrad_losses(self, edge_logit, maskgrad_prob):
        """Compute maskgrad loss."""
        edge_logit = resize(
            input=edge_logit,
            size=maskgrad_prob.shape[2:],  # (b, 19, h, w)
            mode="bilinear",
            align_corners=self.align_corners,
        )
        return {
            self.mask_edge_loss.loss_name: self.mask_edge_loss(
                edge_logit=edge_logit,
                maskgrad=maskgrad_prob,
            )
        }

    def forward_train(
        self, inputs, img_metas, gt_semantic_seg, gt_semantic_edge, train_cfg
    ):
        if self.mask_grad:
            seg_logit, edge_logit, maskgrad_edge = self(inputs)
        else:
            seg_logit, edge_logit = self(inputs)

        # assume that there are always seg_losses
        losses = self.seg_losses(
            seg_logit=seg_logit,
            seg_label=gt_semantic_seg,
        )
        if self.loss_edge is not None:
            losses.update(
                self.loss_edge(
                    edge_logit=edge_logit,
                    maskgrad_prob=maskgrad_edge,
                )
            )
        if self.loss_joint is not None:
            # FIXME: multilabel edges are not supported (buggy)
            losses.update(
                self.joint_losses(
                    seg_logit=seg_logit,
                    edge_logit=edge_logit,
                    seg_label=gt_semantic_seg,
                    edge_label=gt_semantic_edge,
                )
            )

        # compute losses for mask edge
        if self.mask_grad:
            if self.mask_edge_loss is not None:
                losses.update(
                    self.maskgrad_losses(
                        edge_logit=edge_logit,
                        maskgrad_prob=maskgrad_edge,
                    )
                )

        return losses

    def forward_test(self, inputs, img_metas, test_cfg, return_edge=False, **kwargs):
        if self.mask_grad:
            seg, edge, _ = self(inputs)
        else:
            seg, edge = self(inputs)
        if return_edge:
            return dict(seg=seg, edge=edge)
        else:
            return dict(seg=seg)


class BaseMaskGradManyEdgeHead(BaseManyEdgeDecodeHead):
    def __init__(
        self,
        mask_grad=False,
        loss_mask_edge=dict(type="DualityLoss", loss_weight=1.0),
        loss_mask_edge_key=None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.mask_grad = mask_grad
        if self.mask_grad:
            assert loss_mask_edge_key is not None
            assert loss_mask_edge_key in self.multilabel_edge_keys
            assert isinstance(loss_mask_edge, dict)
            self.mask_edge_loss = build_edge_loss(loss_mask_edge)
            self.loss_mask_edge_key = loss_mask_edge_key

    @force_fp32(apply_to=("edge_logit", "maskgrad_prob"))
    def maskgrad_losses(self, edge_logit, maskgrad_prob):
        """Compute maskgrad loss."""
        edge_logit = resize(
            input=edge_logit,
            size=maskgrad_prob.shape[2:],  # (b, 19, h, w)
            mode="bilinear",
            align_corners=self.align_corners,
        )
        return {
            self.mask_edge_loss.loss_name: self.mask_edge_loss(
                edge_logit=edge_logit,
                maskgrad=maskgrad_prob,
            )
        }

    def forward_train(
        self, inputs, img_metas, gt_semantic_seg, gt_semantic_edge, train_cfg
    ):
        if self.mask_grad:
            seg_logits, binary_edge_logits, semantic_edge_logits, mask_edge = self(
                inputs
            )
        else:
            seg_logits, binary_edge_logits, semantic_edge_logits = self(inputs)

        # assume that there are always seg_losses
        losses = self.seg_losses(
            seg_logit=seg_logits,
            seg_label=gt_semantic_seg,
        )
        if self.loss_binary_edge is not None:
            losses.update(
                self.binary_edge_losses(
                    edge_logit=binary_edge_logits,
                    edge_label=gt_semantic_edge,
                )
            )
        if self.loss_multilabel_edge is not None:
            losses.update(
                self.multilabel_edge_losses(
                    edge_logit=semantic_edge_logits,
                    edge_label=gt_semantic_edge,
                )
            )
        if self.loss_joint is not None:
            edge_logit = semantic_edge_logits.get(self.loss_joint_edge_key, None)
            if edge_logit is None:
                edge_logit = binary_edge_logits.get(self.loss_joint_edge_key, None)
            assert edge_logit is not None
            losses.update(
                self.joint_losses(
                    seg_logit=seg_logits,
                    edge_logit=edge_logit,
                    seg_label=gt_semantic_seg,
                    edge_label=gt_semantic_edge,
                )
            )

        # compute losses for mask edge
        if self.mask_grad:
            edge_logit = semantic_edge_logits[self.loss_mask_edge_key]
            losses.update(
                self.maskgrad_losses(
                    edge_logit=edge_logit,
                    maskgrad_prob=mask_edge,
                )
            )

        return losses

    def forward_test(self, inputs, img_metas, test_cfg, return_edge=False, **kwargs):
        if self.mask_grad:
            seg, binary_edge, semantic_edge, _ = self(inputs)
        else:
            seg, binary_edge, semantic_edge = self(inputs)
        if return_edge:
            edge = None
            if isinstance(binary_edge, dict):
                edge = binary_edge.get(self.edge_key, None)
            if isinstance(semantic_edge, dict):
                edge = semantic_edge.get(self.edge_key, None)
            assert edge is not None, f"could not find {self.edge_key}"
            return dict(seg=seg, edge=edge)
        else:
            return dict(seg=seg)



#!/usr/bin/env python3

# TODO: refactor and validate

import numpy as np

import torch
import torch.nn.functional as F

# from torchvision.transforms.functional import pad


def calc_pad_same(in_siz, out_siz, stride, ksize):
    """Calculate same padding width.
    Args:
        ksize: kernel size [I, J].
    Returns:
        pad_: Actual padding width.
    """
    return (out_siz - 1) * stride + ksize - in_siz


def conv2d_same(input, kernel, groups, bias=None, stride=1, padding=0, dilation=1):
    n, c, h, w = input.shape
    kout, ki_c_g, kh, kw = kernel.shape
    pw = calc_pad_same(w, w, 1, kw)
    ph = calc_pad_same(h, h, 1, kh)
    pw_l = pw // 2
    pw_r = pw - pw_l
    ph_t = ph // 2
    ph_b = ph - ph_t

    input_ = F.pad(input, (pw_l, pw_r, ph_t, ph_b))
    result = F.conv2d(
        input_,
        kernel,
        bias=bias,
        stride=stride,
        padding=padding,
        dilation=dilation,
        groups=groups,
    )
    assert result.shape == input.shape
    return result


def gradient_central_diff(input):
    # https://github.com/nv-tlabs/GSCNN/issues/16
    # return input, input  # FIXME: typo?

    n, c, h, w = input.shape
    device = input.device

    # pytorch implements correlation instead of conv
    kernel = [[1, 0, -1]]
    kernel_t = 0.5 * torch.Tensor(kernel) * -1.0
    kernel_t = kernel_t.to(device)

    x = conv2d_same(
        input,
        kernel_t.unsqueeze(0).unsqueeze(0).repeat([c, 1, 1, 1]),
        c,
    )
    y = conv2d_same(
        input,
        kernel_t.t().unsqueeze(0).unsqueeze(0).repeat([c, 1, 1, 1]),
        c,
    )
    return x, y


def compute_single_sided_diferences(o_x, o_y, input):
    # n,c,h,w
    # input = input.clone()
    o_y[:, :, 0, :] = input[:, :, 1, :].clone() - input[:, :, 0, :].clone()
    o_x[:, :, :, 0] = input[:, :, :, 1].clone() - input[:, :, :, 0].clone()
    # --
    o_y[:, :, -1, :] = input[:, :, -1, :].clone() - input[:, :, -2, :].clone()
    o_x[:, :, :, -1] = input[:, :, :, -1].clone() - input[:, :, :, -2].clone()
    return o_x, o_y


def numerical_gradients_2d(input):
    """
    numerical gradients implementation over batches using torch group conv operator.
    the single sided differences are re-computed later.
    it matches np.gradient(image) with the difference than here output=x,y for an image while there output=y,x

    Args:
        input: N,C,H,W
    Returns:
        X,Y
    """
    n, c, h, w = input.shape
    assert h > 1 and w > 1
    x, y = gradient_central_diff(input)
    return x, y


def convTri(input, r):
    """
    Convolves an image by a 2D triangle filter (the 1D triangle filter f is
    [1:r r+1 r:-1:1]/(r+1)^2, the 2D version is simply conv2(f,f'))

    Args:
        input:
        r: integer filter radius

    ref: https://github.com/pdollar/toolbox/blob/master/channels/convTri.m
    """
    if r <= 1:
        raise ValueError()
    n, c, h, w = input.shape
    device = input.device

    # return input  # FIXME: typo?

    f = list(range(1, r + 1)) + [r + 1] + list(reversed(range(1, r + 1)))
    kernel = torch.Tensor([f]) / (r + 1) ** 2
    kernel = kernel.to(device)

    # padding w
    input_ = F.pad(input, (1, 1, 0, 0), mode="replicate")
    input_ = F.pad(input_, (r, r, 0, 0), mode="reflect")
    input_ = [input_[:, :, :, :r], input, input_[:, :, :, -r:]]
    input_ = torch.cat(input_, 3)
    t = input_

    # padding h
    input_ = F.pad(input_, (0, 0, 1, 1), mode="replicate")
    input_ = F.pad(input_, (0, 0, r, r), mode="reflect")
    input_ = [input_[:, :, :r, :], t, input_[:, :, -r:, :]]
    input_ = torch.cat(input_, 2)

    output = F.conv2d(
        input_,
        kernel.unsqueeze(0).unsqueeze(0).repeat([c, 1, 1, 1]),
        padding=0,
        groups=c,
    )
    output = F.conv2d(
        output,
        kernel.t().unsqueeze(0).unsqueeze(0).repeat([c, 1, 1, 1]),
        padding=0,
        groups=c,
    )
    return output


def compute_normal(E):
    # NOTE: is this the same normal used in 'STEAL'?

    if torch.sum(torch.isnan(E)) != 0:
        print("nans found here")
        import ipdb

        ipdb.set_trace()

    E_ = convTri(E, 4)
    Ox, Oy = numerical_gradients_2d(E_)
    Oxx, _ = numerical_gradients_2d(Ox)
    Oxy, Oyy = numerical_gradients_2d(Oy)

    aa = Oyy * torch.sign(-(Oxy + 1e-5)) / (Oxx + 1e-5)
    t = torch.atan(aa)
    O = torch.remainder(t, np.pi)

    if torch.sum(torch.isnan(O)) != 0:
        print("nans found here")
        import ipdb

        ipdb.set_trace()

    return O


def compute_normal_2(E):
    # output Oyy and Oxx... for what?

    if torch.sum(torch.isnan(E)) != 0:
        print("nans found here")
        import ipdb

        ipdb.set_trace()

    E_ = convTri(E, 4)
    Ox, Oy = numerical_gradients_2d(E_)
    Oxx, _ = numerical_gradients_2d(Ox)
    Oxy, Oyy = numerical_gradients_2d(Oy)

    aa = Oyy * torch.sign(-(Oxy + 1e-5)) / (Oxx + 1e-5)
    t = torch.atan(aa)
    O = torch.remainder(t, np.pi)

    if torch.sum(torch.isnan(O)) != 0:
        print("nans found here")
        import ipdb

        ipdb.set_trace()

    return O, (Oyy, Oxx)


def compute_grad_mag(E, eps=1e-6):

    # FIXME: what is `mag`?
    # magnitude?

    if torch.sum(torch.isnan(E)) != 0:
        print("nans found here")
        import ipdb

        ipdb.set_trace()

    # FIXME: input of convTri is sometimes NaN
    E_ = convTri(E, 4)
    Ox, Oy = numerical_gradients_2d(E_)
    # mag = torch.sqrt(torch.mul(Ox, Ox) + torch.mul(Oy, Oy) + 1e-6)
    # mag = mag / mag.max()  # divide by zero error

    mag = torch.sqrt(torch.mul(Ox, Ox) + torch.mul(Oy, Oy) + eps)
    mag = torch.div(mag, mag.max() + eps)

    if torch.sum(torch.isnan(mag)) != 0:
        print("nans found here")
        import ipdb

        ipdb.set_trace()

    assert torch.sum(torch.isinf(mag)) == 0, f"{mag}"

    return mag



#!/usr/bin/env python3

"""3StreamHead

Semantic segmentation, Multi-label edge detection, binary edge detection,
all-in-one (ASPPHead)
"""

import torch
import torch.nn as nn

from mmcv.cnn import ConvModule, DepthwiseSeparableConvModule, build_activation_layer
from mmseg.ops import resize
from mmseg.models.decode_heads.sep_aspp_head import DepthwiseSeparableASPPModule
from blette.models.utils import LocationAdaptiveLearner, SideConv

from potato.models.builder import HEADS
from potato.models.utils import maskgrad
from potato.ops.canny_edge import UnNormalizedCanny
from potato.ops.torch_canny_edge import TorchUnNormalizedSoftCanny
from .maskgrad_decode_head import BaseMaskGradManyEdgeHead
from ..utils import ResNetGSCNNShapeStream


@HEADS.register_module()
class ThreeStreamASPPHead(BaseMaskGradManyEdgeHead):
    def __init__(
        self,
        dilations=(1, 6, 12, 18),
        c1_in_channels=256,
        c1_channels=48,
        edge_key="fuse",
        log_edge_keys=["fuse", "binary"],
        binary_edge_keys=["binary"],
        multilabel_edge_keys=["fuse", "side5"],
        loss_binary_edge=dict(type="BinaryEdgeLoss", loss_weight=1.0 * 20),
        loss_multilabel_edge=dict(type="MultiLabelEdgeLoss", loss_weight=1.0),
        cv2_canny=False,
        mask_grad=False,
        loss_mask_edge=dict(type="MaskGradientCrossEntropyLoss", loss_weight=1.0),
        **kwargs,
    ):
        super().__init__(
            input_transform="multiple_select",
            edge_key=edge_key,
            log_edge_keys=log_edge_keys,
            binary_edge_keys=binary_edge_keys,
            multilabel_edge_keys=multilabel_edge_keys,
            loss_binary_edge=loss_binary_edge,
            loss_multilabel_edge=loss_multilabel_edge,
            mask_grad=mask_grad,
            loss_mask_edge=loss_mask_edge,
            **kwargs,
        )

        _interp = "bilinear"

        # 1. Multi-label edge stream
        self.side1 = SideConv(
            in_channels=self.in_channels[0],
            out_channels=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side2 = SideConv(
            in_channels=self.in_channels[1],
            out_channels=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side3 = SideConv(
            in_channels=self.in_channels[2],
            out_channels=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side5 = SideConv(
            in_channels=self.in_channels[4],
            out_channels=self.num_classes,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.side5_w = SideConv(
            in_channels=self.in_channels[4],
            out_channels=self.num_classes * 4,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )
        self.ada_learner = LocationAdaptiveLearner(
            self.num_classes * 4,
            self.num_classes * 4,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        # 2. Shape stream
        self.shape_stream = ResNetGSCNNShapeStream(
            in_channels=self.in_channels,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
            interpolation=_interp,
            align_corners=self.align_corners,
        )

        self.cw = ConvModule(
            in_channels=2,  # shape + canny
            out_channels=1,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        self.aspp_edge_conv = ConvModule(
            in_channels=1,
            out_channels=self.channels,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        # 3. Fusion
        self.image_pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(
                self.in_channels[4],
                self.channels,
                1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )
        self.aspp_modules = DepthwiseSeparableASPPModule(
            in_channels=self.in_channels[4],
            channels=self.channels,
            dilations=dilations,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.c1_bottleneck = ConvModule(
            in_channels=c1_in_channels,
            out_channels=c1_channels,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.bottleneck = ConvModule(
            (len(dilations) + 2) * self.channels,
            self.channels,
            kernel_size=3,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.sep_bottleneck = nn.Sequential(
            DepthwiseSeparableConvModule(
                in_channels=self.channels + c1_channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
            DepthwiseSeparableConvModule(
                in_channels=self.channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )

        self.activate = build_activation_layer(self.act_cfg)

        if cv2_canny:
            # cv2, fast, non-differentiable
            self.canny = UnNormalizedCanny()
        else:
            # torch, a bit slower, differentiable, no hysteresis
            self.canny = TorchUnNormalizedSoftCanny()

        if self.mask_grad:
            self.grouped_edge_conv = nn.Conv2d(
                in_channels=self.num_classes * 2,
                out_channels=self.num_classes,
                kernel_size=1,
                groups=self.num_classes,
            )

    def forward(self, inputs):
        """Forward function"""

        # x = self._transform_inputs(inputs)  # out: list
        x = [i for i in inputs]
        # [stem, layer1, layer2, layer3, layer4, input_image]
        # h: 1/2, 1/4, 1/8, 1/8, 1/8, 1

        assert isinstance(x, list)
        assert len(x) == 6

        bs, c, h, w = x[-1].shape
        resize_to = (h, w)  # TODO: might be too large

        # Shape stream
        binary_edge = self.shape_stream(x)
        canny = self.canny(x[-1])
        edge_cat = torch.cat((binary_edge, canny), dim=1)
        edge_feat = self.cw(edge_cat)  # do we need sigmoid here?
        edge_feat = resize(
            edge_feat,
            size=x[4].size()[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )

        # Semantic Edge Stream
        side1 = self.side1(x[0], resize_to)  # (B, 1, H, W)
        side2 = self.side2(x[1], resize_to)  # (B, 1, H, W)
        side3 = self.side3(x[2], resize_to)  # (B, 1, H, W)
        side5 = self.side5(x[4], resize_to)  # (B, 19, H, W)
        side5_w = self.side5_w(x[4], resize_to)  # (B, 19*4, H, W)
        fused_edges = self.ada_learner([side1, side2, side3, side5, side5_w])

        # ASPP
        aspp_outs = [
            resize(
                self.image_pool(x[4]),
                size=x[4].size()[2:],
                mode="bilinear",
                align_corners=self.align_corners,
            )
        ]
        aspp_outs.append(self.aspp_edge_conv(edge_feat))
        aspp_outs.extend(self.aspp_modules(x[4]))
        aspp_outs = torch.cat(aspp_outs, dim=1)  # fusing here, might not do much
        output = self.bottleneck(aspp_outs)
        c1_output = self.c1_bottleneck(x[1])
        # fuse_feat = resize(
        #     fused_edges,
        #     size=c1_output.shape[2:],
        #     mode="bilinear",
        #     align_corners=self.align_corners,
        # )
        # print("edge_feat", edge_feat.shape)  # (b, 1, 64, 128) resolution too small...?
        # print("fused", fuse_feat.shape)  # (b, 19, 64, 128)
        output = resize(
            input=output,
            size=c1_output.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )
        output = torch.cat([output, c1_output], dim=1)
        output = self.sep_bottleneck(output)
        output = self.cls_seg(output)

        if self.mask_grad:
            # calculate mask gradient
            seg_out = resize(
                output,  # we want logits
                size=resize_to,
                mode="bilinear",
                align_corners=self.align_corners,
            )
            mask_edge = maskgrad(
                seg_map=seg_out,
                sensitivity=100,
            )
            # sliced concatenation and k-grouped convolution
            hold = []
            for c in range(fused_edges.size(1)):
                hold.append(fused_edges[:, c, ...].unsqueeze(1))
                hold.append(mask_edge[:, c, ...].unsqueeze(1))
            edge = torch.cat(hold, dim=1)
            edge = self.grouped_edge_conv(edge)  # logits
            binary_edge = dict(binary=binary_edge)
            semantic_edge = dict(fuse=edge, side5=side5)
            return output, binary_edge, semantic_edge, mask_edge
        else:
            binary_edge = dict(binary=binary_edge)
            semantic_edge = dict(fuse=fused_edges, side5=side5)
            return output, binary_edge, semantic_edge



#!/usr/bin/env python3

from .inference import inference_segmentor, init_segmentor, show_seg_result_pyplot
from .test import (
    multi_gpu_seg_test,
    single_gpu_seg_test,
)
from .train import get_root_logger, init_random_seed, set_random_seed, train_seg_det

__all__ = [
    "get_root_logger",
    "set_random_seed",
    "train_seg_det",
    "init_segmentor",
    "inference_segmentor",
    "multi_gpu_seg_test",
    "single_gpu_seg_test",
    "show_seg_result_pyplot",
    "init_random_seed",
]



#!/usr/bin/env python3

from setuptools import find_packages, setup


def readme():
    with open("README.md", encoding="utf-8") as f:
        content = f.read()
    return content


version_file = "potato/version.py"


def get_version():
    with open(version_file, "r") as f:
        exec(compile(f.read(), version_file, "exec"))
    return locals()["__version__"]


def parse_requirements(fname="requirements.txt", with_version=True):
    """Parse the package dependencies listed in a requirements file but strips
    specific versioning information.
    Args:
        fname (str): path to requirements file
        with_version (bool, default=False): if True include version specs
    Returns:
        List[str]: list of requirements items
    CommandLine:
        python -c "import setup; print(setup.parse_requirements())"
    """
    import re
    import sys  # noqa: F811
    from os.path import exists

    require_fpath = fname

    def parse_line(line):
        """Parse information from a line in a requirements text file."""
        if line.startswith("-r "):
            # Allow specifying requirements in other files
            target = line.split(" ")[1]
            for info in parse_require_file(target):
                yield info
        else:
            info = {"line": line}
            if line.startswith("-e "):
                info["package"] = line.split("#egg=")[1]
            else:
                # Remove versioning from the package
                pat = "(" + "|".join([">=", "==", ">"]) + ")"
                parts = re.split(pat, line, maxsplit=1)
                parts = [p.strip() for p in parts]

                info["package"] = parts[0]
                if len(parts) > 1:
                    op, rest = parts[1:]
                    if ";" in rest:
                        # Handle platform specific dependencies
                        # http://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-platform-specific-dependencies
                        version, platform_deps = map(str.strip, rest.split(";"))
                        info["platform_deps"] = platform_deps
                    else:
                        version = rest  # NOQA
                    info["version"] = (op, version)
            yield info

    def parse_require_file(fpath):
        with open(fpath, "r") as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith("#"):
                    for info in parse_line(line):
                        yield info

    def gen_packages_items():
        if exists(require_fpath):
            for info in parse_require_file(require_fpath):
                parts = [info["package"]]
                if with_version and "version" in info:
                    parts.extend(info["version"])
                if not sys.version.startswith("3.4"):
                    # apparently package_deps are broken in 3.4
                    platform_deps = info.get("platform_deps")
                    if platform_deps is not None:
                        parts.append(";" + platform_deps)
                item = "".join(parts)
                yield item

    packages = list(gen_packages_items())
    return packages


setup(
    name="potato-cv",
    version=get_version(),
    description="JSB Framework",
    long_description=readme(),
    long_description_content_type="text/markdown",
    author="Haruya Ishikawa",
    author_email="",
    keywords="computer vision, deep learning",
    url="http://github.com/haruishi43/potato",
    packages=find_packages(exclude=("configs", "tools", "demo", "tests")),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    license="Apache License 2.0",
    install_requires=parse_requirements("requirements.txt"),
    ext_modules=[],
    zip_safe=False,
)



# Copyright (c) OpenMMLab. All rights reserved.
from .collect_env import collect_env
from .logger import get_root_logger
from .misc import find_latest_checkpoint
from .set_env import setup_multi_processes

__all__ = [
    "get_root_logger",
    "collect_env",
    "find_latest_checkpoint",
    "setup_multi_processes",
]



#!/usr/bin/env python3

from .accuracy import Accuracy, accuracy, edge_accuracy
from .cross_entropy_loss import (
    CrossEntropyLoss,
    binary_cross_entropy,
    cross_entropy,
    mask_cross_entropy,
)
from .utils import reduce_loss, weight_reduce_loss, weighted_loss

from .old_losses import (
    EdgeAttentionLoss,
    ImageBasedCrossEntropyLoss2d,
)
from .dual_task_loss import DualTaskLoss
from .mask_gradient_loss import DualityLoss, MaskGradientCrossEntropyLoss
from .boundary_loss import BoundaryLoss

__all__ = [
    "Accuracy",
    "accuracy",
    "edge_accuracy",
    "CrossEntropyLoss",
    "cross_entropy",
    "binary_cross_entropy",
    "mask_cross_entropy",
    "reduce_loss",
    "weight_reduce_loss",
    "weighted_loss",
    "EdgeAttentionLoss",
    "ImageBasedCrossEntropyLoss2d",
    "DualTaskLoss",
    "MaskGradientCrossEntropyLoss",
    "DualityLoss",
    "BoundaryLoss",
]



#!/usr/bin/env python3

import torch
import torch.nn as nn
import torch.nn.functional as F

from mmcv.cnn.bricks import build_activation_layer, build_norm_layer
from mmseg.models.decode_heads.decode_head import BaseDecodeHead

from ..builder import HEADS


class DenseASPPBlock(nn.Sequential):
    """ConvNet block for building DenseASPP."""

    def __init__(
        self,
        in_channels,
        mid_channels,
        out_channels,
        dilation_rate,
        drop_rate,
        bn_start=True,
        norm_cfg=dict(type="BN"),
        act_cfg=dict(type="ReLU"),
    ):
        super(DenseASPPBlock, self).__init__()

        # force momentum on Norm (Assuming BN or SyncBN)
        norm_cfg.update({"momentum": 0.0003})

        if bn_start:
            self.add_module("norm1", build_norm_layer(norm_cfg, in_channels)[1])

        self.add_module("relu1", build_activation_layer(act_cfg))
        self.add_module(
            "conv1",
            nn.Conv2d(
                in_channels=in_channels, out_channels=mid_channels, kernel_size=1
            ),
        )

        self.add_module("norm2", build_norm_layer(norm_cfg, mid_channels)[1])
        self.add_module("relu2", build_activation_layer(act_cfg))
        self.add_module(
            "conv2",
            nn.Conv2d(
                in_channels=mid_channels,
                out_channels=out_channels,
                kernel_size=3,
                dilation=dilation_rate,
                padding=dilation_rate,
            ),
        )

        self.drop_rate = drop_rate

    def forward(self, input):
        feature = super(DenseASPPBlock, self).forward(input)
        if self.drop_rate > 0:
            feature = F.dropout2d(feature, p=self.drop_rate, training=self.training)
        return feature


class DenseASPP(nn.Module):
    def __init__(
        self,
        in_channels,
        mid_channels,
        out_channels,
        dilation_rates=(3, 6, 12, 18, 24),
        drop_rate=0,
    ):
        super(DenseASPP, self).__init__()

        self.aspps = nn.ModuleList()
        for i, dil in enumerate(dilation_rates):
            self.aspps.append(
                DenseASPPBlock(
                    in_channels=in_channels + out_channels * i,
                    mid_channels=mid_channels,
                    out_channels=out_channels,
                    dilation_rate=dil,
                    drop_rate=drop_rate,
                    bn_start=False if i == 0 else True,
                )
            )

    def forward(self, x):
        for aspp in self.aspps:
            f = aspp(x)
            x = torch.cat((x, f), dim=1)
        return x


@HEADS.register_module()
class DenseASPPHead(BaseDecodeHead):
    """DenseASPP.

    This head is the implementation of
    `DenseASPP <https://ieeexplore.ieee.org/document/8578486>`

    Args:
        in_channels (int): input channels for DenseASPP (refer to the backbone)
        channels (int): output channels for each DenseASPPBlock
        mid_channels (int): output channels for the first conv in DenseASPPBlock
        dilations (tuple(int)): dilation_ratio for each DenseASPPBlock
        aspp_dropout_ratio (float): dropout ratio for DenseASPPBlock
    """

    def __init__(
        self,
        in_channels,
        channels,
        mid_channels,
        dilations=(3, 6, 12, 18, 24),
        aspp_dropout_ratio=0.1,
        **kwargs,
    ):
        # pre-calc output channels for final seg
        final_channels = in_channels + len(dilations) * channels
        super(DenseASPPHead, self).__init__(
            in_channels=in_channels,
            channels=final_channels,
            **kwargs,
        )

        self.dense_aspp = DenseASPP(
            in_channels=self.in_channels,
            mid_channels=mid_channels,
            out_channels=channels,
            dilation_rates=dilations,
            drop_rate=aspp_dropout_ratio,
        )

    def forward(self, inputs):
        x = self._transform_inputs(inputs)
        x = self.dense_aspp(x)
        output = self.cls_seg(x)
        return output



#!/usr/bin/env python3

import math

import torch
import torch.nn.functional as F


def extract_image_patches(x, patch_size=(3, 3)):
    b, c, h, w = x.shape
    patch_h = math.ceil(h / patch_size[0])
    patch_w = math.ceil(w / patch_size[1])
    pad_h = patch_size[0] * patch_h - h
    pad_w = patch_size[1] * patch_w - w
    # pad: left, right, top, bottom
    x = F.pad(x, (pad_w // 2, pad_w - pad_w // 2, pad_h // 2, pad_h - pad_h // 2))

    # Extract patches
    patches = x.unfold(2, patch_h, patch_h).unfold(3, patch_w, patch_w)
    patches = patches.permute(0, 1, 2, 3, 4, 5).contiguous()

    return patches.view(b, -1, c, patches.shape[-2], patches.shape[-1])


def patch_wise_pool(x, patch_size=(3, 3)):
    b, p, c, h, w = x.shape

    vs = []
    for i in range(p):
        vs.append(
            F.avg_pool2d(x[:, i, ...], (h, w)).unsqueeze(-1).unsqueeze(-1)
        )  # (b, c)

    f = torch.stack(vs, dim=-3)
    f = f.view(b, c, *patch_size)

    return f  # (b, c, p, p)



#!/usr/bin/env python3

import torch
import torch.nn as nn

from mmcv.cnn import ConvModule, DepthwiseSeparableConvModule
from mmseg.ops import resize
from mmseg.models.decode_heads.decode_head import BaseDecodeHead
from mmseg.models.decode_heads.sep_aspp_head import DepthwiseSeparableASPPModule

from potato.models.builder import HEADS
from potato.ops.canny_edge import UnNormalizedCanny
from potato.ops.torch_canny_edge import TorchUnNormalizedSoftCanny


@HEADS.register_module()
class GradASPPHead(BaseDecodeHead):
    def __init__(
        self,
        dilations=(1, 6, 12, 18),
        c1_in_channels=256,
        c1_channels=48,
        cv2_canny=False,
        **kwargs,
    ):
        super().__init__(input_transform="multiple_select", **kwargs)

        self.image_pool = nn.Sequential(
            nn.AdaptiveAvgPool2d(1),
            ConvModule(
                self.in_channels[4],
                self.channels,
                1,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )
        self.aspp_modules = DepthwiseSeparableASPPModule(
            in_channels=self.in_channels[4],
            channels=self.channels,
            dilations=dilations,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.c1_bottleneck = ConvModule(
            in_channels=c1_in_channels,
            out_channels=c1_channels,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.bottleneck = ConvModule(
            (len(dilations) + 1) * self.channels + 1,  # add edge
            self.channels,
            3,
            padding=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )
        self.sep_bottleneck = nn.Sequential(
            DepthwiseSeparableConvModule(
                in_channels=self.channels + c1_channels + self.channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
            DepthwiseSeparableConvModule(
                in_channels=self.channels,
                out_channels=self.channels,
                kernel_size=3,
                padding=1,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg,
            ),
        )

        self.cw = ConvModule(
            in_channels=1,
            out_channels=1,
            kernel_size=1,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=None,  # using sigmoid
        )
        self.edge_conv = ConvModule(
            in_channels=1,
            out_channels=self.channels,
            kernel_size=1,
            padding=0,
            conv_cfg=self.conv_cfg,
            norm_cfg=self.norm_cfg,
            act_cfg=self.act_cfg,
        )

        if cv2_canny:
            # cv2, fast, non-differentiable
            self.canny = UnNormalizedCanny()
        else:
            # torch, a bit slower, differentiable, no hysteresis
            self.canny = TorchUnNormalizedSoftCanny()

    def forward(self, inputs):
        """Forward function"""

        # x = self._transform_inputs(inputs)  # out: list
        x = [i for i in inputs]
        # [stem, layer1, layer2, layer3, layer4, input_image]
        # h: 1/2, 1/4, 1/8, 1/8, 1/8, 1

        inp_bot = x[1]
        inp = x[4]

        # Canny Edge
        # FIXME: change this to image gradient
        canny = self.canny(x[-1])
        canny_feat = torch.sigmoid(self.cw(canny))  # remove sigmoid?
        deep_canny_feat = resize(
            canny_feat,
            size=inp.size()[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )

        aspp_outs = [
            resize(
                self.image_pool(inp),
                size=inp.size()[2:],
                mode="bilinear",
                align_corners=self.align_corners,
            )
        ]
        aspp_outs.append(deep_canny_feat)  # 1. introduce canny features
        aspp_outs.extend(self.aspp_modules(inp))
        aspp_outs = torch.cat(aspp_outs, dim=1)
        output = self.bottleneck(aspp_outs)
        c1_output = self.c1_bottleneck(inp_bot)
        output = resize(
            input=output,
            size=c1_output.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )
        shallower_canny_feat = resize(
            input=canny_feat,
            size=c1_output.shape[2:],
            mode="bilinear",
            align_corners=self.align_corners,
        )
        sem_edge_feat = self.edge_conv(shallower_canny_feat)
        # 2. introduce canny features into shallower head
        output = torch.cat([output, c1_output, sem_edge_feat], dim=1)
        output = self.sep_bottleneck(output)
        output = self.cls_seg(output)

        return output



#!/usr/bin/env python3

"""Canny Edge

- Implement pytorch version of the algorithm
- [Canny Edge Pytorch](https://github.com/DCurro/CannyEdgePytorch)
- Already implemented by kornia, but non-differentiable, and very slow
"""

import math
from typing import Tuple

import torch
import torch.nn.functional as F

from kornia.color import rgb_to_grayscale
from kornia.filters import canny
from kornia.filters.gaussian import gaussian_blur2d
from kornia.filters.kernels import get_canny_nms_kernel  # , get_hysteresis_kernel
from kornia.filters.sobel import spatial_gradient

from .transforms import UnNormalize

__all__ = [
    "canny",
    "soft_canny",
    "TorchUnNormalizedCanny",
    "TorchUnNormalizedSoftCanny",
]


def soft_canny(
    input,
    sharpness: float = 1.0,
    low_threshold: float = 0.1,
    high_threshold: float = 0.2,
    kernel_size: Tuple[int, int] = (5, 5),
    sigma: Tuple[float, float] = (1, 1),
    hysteresis: bool = True,
    eps: float = 1e-6,
):
    if not isinstance(input, torch.Tensor):
        raise TypeError(f"Input type is not a torch.Tensor. Got {type(input)}")

    if not len(input.shape) == 4:
        raise ValueError(f"Invalid input shape, we expect BxCxHxW. Got: {input.shape}")

    if low_threshold > high_threshold:
        raise ValueError(
            "Invalid input thresholds. low_threshold should be smaller than the high_threshold. Got: {}>{}".format(
                low_threshold, high_threshold
            )
        )

    # TODO: add warnings for hysteresis

    if low_threshold < 0 and low_threshold > 1:
        raise ValueError(
            f"Invalid input threshold. low_threshold should be in range (0,1). Got: {low_threshold}"
        )

    if high_threshold < 0 and high_threshold > 1:
        raise ValueError(
            f"Invalid input threshold. high_threshold should be in range (0,1). Got: {high_threshold}"
        )

    device: torch.device = input.device
    dtype: torch.dtype = input.dtype

    # To Grayscale
    if input.shape[1] == 3:
        input = rgb_to_grayscale(input)

    # Gaussian filter
    blurred: torch.Tensor = gaussian_blur2d(input, kernel_size, sigma)

    # Compute the gradients
    gradients: torch.Tensor = spatial_gradient(blurred, normalized=False)

    # Unpack the edges
    gx: torch.Tensor = gradients[:, :, 0]
    gy: torch.Tensor = gradients[:, :, 1]

    # Compute gradient magnitude and angle
    magnitude: torch.Tensor = torch.sqrt(gx * gx + gy * gy + eps)
    angle: torch.Tensor = torch.atan2(gy, gx)

    # Radians to Degrees
    angle = 180.0 * angle / math.pi

    # Round angle to the nearest 45 degree
    angle = torch.round(angle / 45) * 45

    # Non-maximal suppression
    nms_kernels: torch.Tensor = get_canny_nms_kernel(device, dtype)
    nms_magnitude: torch.Tensor = F.conv2d(
        magnitude, nms_kernels, padding=nms_kernels.shape[-1] // 2
    )

    # Get the indices for both directions
    positive_idx: torch.Tensor = (angle / 45) % 8
    positive_idx = positive_idx.long()

    negative_idx: torch.Tensor = ((angle / 45) + 4) % 8
    negative_idx = negative_idx.long()

    # Apply the non-maximum suppresion to the different directions
    channel_select_filtered_positive: torch.Tensor = torch.gather(
        nms_magnitude, 1, positive_idx
    )
    channel_select_filtered_negative: torch.Tensor = torch.gather(
        nms_magnitude, 1, negative_idx
    )

    channel_select_filtered: torch.Tensor = torch.stack(
        [channel_select_filtered_positive, channel_select_filtered_negative], 1
    )

    # is_max: torch.Tensor = channel_select_filtered.min(dim=1)[0] > 0.0
    # relaxed thresholding
    is_max: torch.Tensor = torch.sigmoid(channel_select_filtered.min(dim=1)[0])

    magnitude = magnitude * is_max

    # Threshold
    edges: torch.Tensor = F.threshold(magnitude, low_threshold, 0.0)

    # low: torch.Tensor = magnitude > low_threshold
    # high: torch.Tensor = magnitude > high_threshold
    # relax thresholding
    low = torch.sigmoid(sharpness * (magnitude - low_threshold))
    high = torch.sigmoid(sharpness * (magnitude - high_threshold))

    edges = low * 0.5 + high * 0.5
    edges = edges.to(dtype)

    # non-differentiable
    # if hysteresis:
    #     edges_old: torch.Tensor = -torch.ones(edges.shape, device=edges.device, dtype=dtype)
    #     hysteresis_kernels: torch.Tensor = get_hysteresis_kernel(device, dtype)

    #     while ((edges_old - edges).abs() != 0).any():
    #         weak: torch.Tensor = (edges == 0.5).float()
    #         strong: torch.Tensor = (edges == 1).float()

    #         hysteresis_magnitude: torch.Tensor = F.conv2d(
    #             edges, hysteresis_kernels, padding=hysteresis_kernels.shape[-1] // 2
    #         )
    #         hysteresis_magnitude = (hysteresis_magnitude == 1).any(1, keepdim=True).to(dtype)
    #         hysteresis_magnitude = hysteresis_magnitude * weak + strong

    #         edges_old = edges.clone()
    #         edges = hysteresis_magnitude + (hysteresis_magnitude == 0) * weak * 0.5

    #     edges = hysteresis_magnitude

    return edges


class TorchUnNormalizedCanny(object):
    def __init__(
        self,
        low_threshold: float = 0.1,
        high_threshold: float = 0.2,
        kernel_size: Tuple[int, int] = (5, 5),
        sigma: Tuple[float, float] = (1, 1),
        hysteresis: bool = True,
        eps: float = 1e-6,
        mean: Tuple[float, float, float] = [123.675, 116.28, 103.53],
        std: Tuple[float, float, float] = [58.395, 57.12, 57.375],
        inplace: bool = False,  # NOTE: default should be false
    ):
        self._low_threshold = low_threshold
        self._high_threshold = high_threshold
        self._kernel_size = kernel_size
        self._sigma = sigma
        self._hysteresis = hysteresis
        self._eps = eps

        self.unnorm = UnNormalize(
            mean=mean,
            std=std,
            inplace=inplace,
        )

    def __call__(self, tensor):

        non_batch = False
        if tensor.ndim == 3:
            # add batch direction
            non_batch = True
            tensor = tensor.unsqueeze(0)
        assert tensor.ndim == 4

        x = self.unnorm(tensor) / 255.0
        mag, edge = canny(
            input=x,
            low_threshold=self._low_threshold,
            high_threshold=self._high_threshold,
            kernel_size=self._kernel_size,
            sigma=self._sigma,
            hysteresis=self._hysteresis,
            eps=self._eps,
        )

        if non_batch:
            # squeeze on
            edge = edge.squeeze(0)

        return edge


class TorchUnNormalizedSoftCanny(TorchUnNormalizedCanny):
    def __init__(
        self,
        sharpness=50.0,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self._sharpness = sharpness

    def __call__(self, tensor):

        non_batch = False
        if tensor.ndim == 3:
            # add batch direction
            non_batch = True
            tensor = tensor.unsqueeze(0)
        assert tensor.ndim == 4

        x = self.unnorm(tensor) / 255.0
        out = soft_canny(
            input=x,
            sharpness=self._sharpness,
            low_threshold=self._low_threshold,
            high_threshold=self._high_threshold,
            kernel_size=self._kernel_size,
            sigma=self._sigma,
            hysteresis=self._hysteresis,
            eps=self._eps,
        )

        if non_batch:
            # squeeze on
            out = out.squeeze(0)

        return out
